"""HRW (Highest Random Weight) load balancing strategy implementation."""

import hashlib
from typing import Any, Optional, Sequence

from naylence.fame.core import FameEnvelope

from .load_balancing_strategy import LoadBalancingStrategy


class HRWLoadBalancingStrategy(LoadBalancingStrategy):
    def __init__(self, *, hash_func=None, sticky_attr: Optional[str] = None):
        self.hash = hash_func or (lambda s: int(hashlib.sha256(s.encode()).hexdigest(), 16))
        self.sticky_attr = sticky_attr

    def choose(self, pool_key: Any, segments: Sequence[str], envelope: FameEnvelope) -> Optional[str]:
        if not segments:
            return None

        # pick salt from sticky_attr if set, else use a random per-envelope salt
        if self.sticky_attr:
            salt = getattr(envelope, self.sticky_attr, None) or envelope.id
        else:
            salt = envelope.id  # random every time → effectively no stickiness

        best = max(segments, key=lambda s: self.hash(f"{s}:{salt}"))
        return best



from typing import Any, Optional, TypeVar

from pydantic import ConfigDict, Field

from naylence.fame.factory import ResourceConfig, ResourceFactory
from naylence.fame.storage.storage_provider import StorageProvider


class StorageProviderConfig(ResourceConfig):
    model_config = ConfigDict(extra="allow")
    type: str = "StorageProvider"
    params: Optional[dict[str, Any]] = Field(default=None, description="Backend-specific kwargs")


C = TypeVar("C", bound=StorageProviderConfig)


class StorageProviderFactory(ResourceFactory[StorageProvider, C]):
    """Pluggable factory, mirrors existing RouteStoreFactory pattern."""

    ...



"""
Factory for creating DefaultSecurityPolicy instances.
"""

from __future__ import annotations

from typing import Any, Optional

from pydantic import Field

from .security_policy import (
    EncryptionConfig,
    SecurityPolicy,
    SecurityPolicyConfig,
    SigningConfig,
)
from .security_policy_factory import SecurityPolicyFactory


class DefaultSecurityPolicyConfig(SecurityPolicyConfig):
    """Configuration for DefaultSecurityPolicy."""

    type: str = "DefaultSecurityPolicy"

    signing: Optional[SigningConfig] = Field(default=None)
    encryption: Optional[EncryptionConfig] = Field(default=None)


class DefaultSecurityPolicyFactory(SecurityPolicyFactory):
    """Factory for creating DefaultSecurityPolicy instances."""

    type: str = "DefaultSecurityPolicy"
    is_default: bool = True

    async def create(
        self,
        config: Optional[DefaultSecurityPolicyConfig | dict[str, Any]] = None,
        **kwargs: Any,
    ) -> SecurityPolicy:
        """Create a DefaultSecurityPolicy instance."""
        from .default_security_policy import DefaultSecurityPolicy

        if config is None:
            config = DefaultSecurityPolicyConfig()
        elif isinstance(config, dict):
            config = DefaultSecurityPolicyConfig(**config)

        encryption = config.encryption
        signing = config.signing

        # Allow runtime overrides from kwargs
        if "encryption" in kwargs:
            encryption = kwargs.pop("encryption")

        if "signing" in kwargs:
            signing = kwargs.pop("signing")

        return DefaultSecurityPolicy(encryption=encryption, signing=signing, **kwargs)



from .flow_controller import FlowController
from .in_memory.in_memory_binding import InMemoryBinding
from .in_memory.in_memory_channel import InMemoryReadWriteChannel

__all__ = [
    "InMemoryBinding",
    "InMemoryReadWriteChannel",
    "FlowController",
]



from typing import Optional, Protocol, runtime_checkable

from naylence.fame.core import Stickiness


@runtime_checkable
class ReplicaStickinessManager(Protocol):
    def offer(self) -> Optional[Stickiness]: ...

    def accept(self, stickiness: Optional[Stickiness]) -> None: ...



"""
Factory for creating BasicAuthorizationPolicy instances.
"""

from __future__ import annotations

from typing import Any

from pydantic import ConfigDict

from naylence.fame.security.auth.policy.authorization_policy import (
    AuthorizationPolicy,
)
from naylence.fame.security.auth.policy.authorization_policy_definition import (
    AuthorizationPolicyDefinition,
)
from naylence.fame.security.auth.policy.authorization_policy_factory import (
    AUTHORIZATION_POLICY_FACTORY_BASE_TYPE,
    AuthorizationPolicyConfig,
    AuthorizationPolicyFactory,
)


class BasicAuthorizationPolicyConfig(AuthorizationPolicyConfig):
    """Configuration for creating a BasicAuthorizationPolicy via factory."""

    model_config = ConfigDict(extra="allow")

    type: str = "BasicAuthorizationPolicy"

    # The policy definition to evaluate - can be dict or Pydantic model
    policy_definition: AuthorizationPolicyDefinition | dict[str, Any] | None = None

    # Whether to log warnings for unknown fields (default: True)
    warn_on_unknown_fields: bool = True


def _normalize_config(
    config: BasicAuthorizationPolicyConfig | dict[str, Any] | None,
) -> dict[str, Any]:
    """Normalize configuration for BasicAuthorizationPolicy."""
    if config is None:
        raise ValueError(
            "BasicAuthorizationPolicyFactory requires a configuration "
            "with a policyDefinition"
        )

    if isinstance(config, BasicAuthorizationPolicyConfig):
        candidate = config.model_dump()
    else:
        candidate = config

    # Support both snake_case and camelCase for policy_definition
    policy_definition = candidate.get("policy_definition") or candidate.get(
        "policyDefinition"
    )

    if not policy_definition:
        raise ValueError(
            "BasicAuthorizationPolicyConfig requires a policyDefinition object"
        )

    # policy_definition can be a dict or already an AuthorizationPolicyDefinition
    if not isinstance(policy_definition, dict | AuthorizationPolicyDefinition):
        raise ValueError(
            "BasicAuthorizationPolicyConfig requires a policyDefinition object"
        )

    # Support both snake_case and camelCase for warn_on_unknown_fields
    warn_on_unknown_fields = candidate.get(
        "warn_on_unknown_fields", candidate.get("warnOnUnknownFields")
    )

    if warn_on_unknown_fields is not None and not isinstance(
        warn_on_unknown_fields, bool
    ):
        raise ValueError("warnOnUnknownFields must be a boolean")

    return {
        "policy_definition": policy_definition,
        "warn_on_unknown_fields": (
            warn_on_unknown_fields if warn_on_unknown_fields is not None else True
        ),
    }


class BasicAuthorizationPolicyFactory(
    AuthorizationPolicyFactory[BasicAuthorizationPolicyConfig]
):
    """Factory for creating BasicAuthorizationPolicy instances."""

    type: str = "BasicAuthorizationPolicy"

    async def create(
        self,
        config: BasicAuthorizationPolicyConfig | dict[str, Any] | None = None,
        **factory_args: Any,
    ) -> AuthorizationPolicy:
        """
        Create a BasicAuthorizationPolicy from the given configuration.

        Args:
            config: Configuration with policyDefinition
            **factory_args: Additional factory arguments (unused)

        Returns:
            The created authorization policy

        Raises:
            ValueError: If configuration is invalid
        """
        normalized = _normalize_config(config)

        # Lazy import to avoid circular dependencies
        from naylence.fame.security.auth.policy.basic_authorization_policy import (
            BasicAuthorizationPolicy,
            BasicAuthorizationPolicyOptions,
        )

        policy_def_raw = normalized["policy_definition"]

        # Convert to AuthorizationPolicyDefinition if needed
        if isinstance(policy_def_raw, dict):
            policy_def = AuthorizationPolicyDefinition.from_dict(policy_def_raw)
        else:
            policy_def = policy_def_raw

        return BasicAuthorizationPolicy(
            BasicAuthorizationPolicyOptions(
                policy_definition=policy_def,
                warn_on_unknown_fields=normalized["warn_on_unknown_fields"],
            )
        )


# Factory metadata for registration
FACTORY_META = {
    "base": AUTHORIZATION_POLICY_FACTORY_BASE_TYPE,
    "key": "BasicAuthorizationPolicy",
}



"""
Security policy interfaces and implementations.
"""

from .default_security_policy import DefaultSecurityPolicy
from .default_security_policy_factory import (
    DefaultSecurityPolicyConfig,
    DefaultSecurityPolicyFactory,
)
from .no_security_policy import NoSecurityPolicy
from .no_security_policy_factory import NoSecurityPolicyConfig, NoSecurityPolicyFactory
from .security_policy import SecurityPolicy, SecurityPolicyConfig
from .security_policy_factory import SecurityPolicyFactory

__all__ = [
    "SecurityPolicy",
    "SecurityPolicyConfig",
    "DefaultSecurityPolicy",
    "NoSecurityPolicy",
    "SecurityPolicyFactory",
    "DefaultSecurityPolicyFactory",
    "DefaultSecurityPolicyConfig",
    "NoSecurityPolicyFactory",
    "NoSecurityPolicyConfig",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional, Protocol, runtime_checkable

from pydantic import BaseModel, ConfigDict, Field

from naylence.fame.core import AuthorizationContext, FameDeliveryContext, FameEnvelope
from naylence.fame.node.node_like import NodeLike

if TYPE_CHECKING:
    from naylence.fame.security.auth.policy.authorization_policy_definition import RuleAction


class RouteAuthorizationResult(BaseModel):
    """
    Route authorization result returned by authorize_route.

    Attributes:
        authorized: Whether the route action is authorized.
        auth_context: The authorization context (if authorized).
        denial_reason: Reason for denial (for internal logging only, not for on-wire disclosure).
        matched_rule: Matched rule ID (for logging/audit).
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    authorized: bool
    auth_context: Optional[AuthorizationContext] = Field(default=None)
    denial_reason: Optional[str] = Field(default=None)
    matched_rule: Optional[str] = Field(default=None)


@runtime_checkable
class Authorizer(Protocol):
    """
    Generic authorization interface supporting multi-phase authentication/authorization.

    This protocol supports both:
    1. Early authentication (token validation from network layer)
    2. Later authorization (envelope-level permission checking including node attach requests)

    The authorize method now accepts the full FameDeliveryContext to enable comprehensive
    authorization decisions based on the complete context including origin, security, and
    authorization information.
    """

    async def authenticate(
        self,
        credentials: str | bytes,
    ) -> Optional[AuthorizationContext]: ...

    async def authorize(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[AuthorizationContext]: ...

    async def authorize_route(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        action: RuleAction,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[RouteAuthorizationResult]:
        """
        Authorizes a routing action after the routing decision has been made.

        This method is called with the explicitly mapped action token from the
        routing decision (ForwardUpstream, ForwardDownstream, ForwardPeer,
        DeliverLocal). It does NOT receive RoutingAction objects to avoid
        coupling authorization logic to routing execution behavior.

        Args:
            node: The node handling the request
            envelope: The FAME envelope being routed
            action: The authorization action token (route-oriented)
            context: Optional delivery context

        Returns:
            RouteAuthorizationResult if implemented, or None to allow
        """
        ...

    def create_reverse_authorization_config(self, node: NodeLike) -> Optional[Any]:
        """
        Create authorization configuration for reverse connections (parent -> child).

        This method allows the authorizer to generate credentials/tokens that can be
        used by a parent node when connecting back to this child node. The returned
        configuration should be a Auth instance suitable for connector configurations.

        Args:
            node: The node that will receive the reverse connection

        Returns:
            Dict containing authorization configuration, or None if reverse auth not supported.
            Authorizers that have a corresponding TokenIssuer can generate appropriate tokens.
        """
        return None  # Default implementation - no reverse auth support



"""
Profile registry for managing factory configuration profiles.

This module provides a registry for factory profiles, allowing factories
to register named configuration templates that can be retrieved and used
at runtime.
"""

from __future__ import annotations

import copy
from dataclasses import dataclass
from typing import Any, Optional

# Type alias for profile configuration
ProfileConfig = dict[str, Any]


@dataclass
class RegisterProfileOptions:
    """Options for registering a profile."""

    allow_override: bool = False
    source: Optional[str] = None


# Internal registry: base_type -> profile_name -> config
_registry: dict[str, dict[str, ProfileConfig]] = {}


def _normalize_key(value: str, label: str) -> str:
    """Normalize and validate a registry key."""
    if not isinstance(value, str):
        raise ValueError(f"{label} must be a non-empty string")

    trimmed = value.strip()
    if not trimmed:
        raise ValueError(f"{label} must be a non-empty string")

    return trimmed


def _clone_config(config: ProfileConfig) -> ProfileConfig:
    """Deep clone a profile configuration."""
    return copy.deepcopy(config)


def register_profile(
    base_type: str,
    name: str,
    config: ProfileConfig,
    options: Optional[RegisterProfileOptions] = None,
) -> None:
    """
    Register a profile configuration for a factory type.

    Args:
        base_type: The factory base type (e.g., 'Authorizer')
        name: The profile name (e.g., 'oauth2', 'noop')
        config: The configuration dictionary for the profile
        options: Optional registration options

    Raises:
        ValueError: If base_type or name are empty
        ValueError: If config is not a dictionary
        ValueError: If profile already exists and allow_override is False
    """
    opts = options or RegisterProfileOptions()
    normalized_base = _normalize_key(base_type, "baseType")
    normalized_name = _normalize_key(name, "profile name")

    if config is None or not isinstance(config, dict):
        raise ValueError(f"Profile '{normalized_name}' config must be an object")

    profiles = _registry.get(normalized_base)
    if profiles is None:
        profiles = {}
        _registry[normalized_base] = profiles

    if normalized_name in profiles and not opts.allow_override:
        source_label = f" ({opts.source})" if opts.source else ""
        raise ValueError(
            f"Profile '{normalized_name}' already registered for "
            f"{normalized_base}{source_label}"
        )

    profiles[normalized_name] = config


def get_profile(base_type: str, name: str) -> Optional[ProfileConfig]:
    """
    Retrieve a profile configuration by type and name.

    Args:
        base_type: The factory base type
        name: The profile name

    Returns:
        A deep copy of the profile configuration, or None if not found
    """
    normalized_base = _normalize_key(base_type, "baseType")
    normalized_name = _normalize_key(name, "profile name")

    profiles = _registry.get(normalized_base)
    if profiles is None:
        return None

    profile = profiles.get(normalized_name)
    return _clone_config(profile) if profile else None


def list_profiles(base_type: str) -> list[str]:
    """
    List all profile names registered for a factory type.

    Args:
        base_type: The factory base type

    Returns:
        List of profile names
    """
    normalized_base = _normalize_key(base_type, "baseType")
    profiles = _registry.get(normalized_base)
    return list(profiles.keys()) if profiles else []


def clear_profiles(base_type: Optional[str] = None) -> None:
    """
    Clear registered profiles.

    Args:
        base_type: If provided, clear only profiles for this type.
                   If None, clear all profiles.
    """
    global _registry

    if base_type is None:
        _registry.clear()
        return

    normalized_base = _normalize_key(base_type, "baseType")
    if normalized_base in _registry:
        del _registry[normalized_base]



from typing import Any, Iterable, Protocol


class KeyProvider(Protocol):
    async def get_key(self, kid: str) -> dict[str, Any]: ...

    async def get_keys_for_path(self, physical_path: str) -> Iterable[dict[str, Any]]: ...


def get_key_provider() -> KeyProvider:
    """Get the key provider instance (returns the key store)."""
    from naylence.fame.security.keys.key_store import get_key_store

    return get_key_store()



"""
Factory for envelope tracker implementations following Fame's ResourceFactory pattern.
"""

from __future__ import annotations

from typing import Any, Optional, TypeVar

from naylence.fame.delivery.delivery_tracker import (
    DeliveryTracker,
    DeliveryTrackerEventHandler,
)
from naylence.fame.delivery.retry_event_handler import RetryEventHandler
from naylence.fame.factory import ResourceConfig, ResourceFactory, create_resource
from naylence.fame.storage.key_value_store import KeyValueStore
from naylence.fame.storage.storage_provider import StorageProvider


class DeliveryTrackerConfig(ResourceConfig):
    """Base configuration for envelope trackers."""

    type: str = "DeliveryTracker"
    namespace: str = "delivery_tracker"


C = TypeVar("C", bound=DeliveryTrackerConfig)


class DeliveryTrackerFactory(ResourceFactory[DeliveryTracker, C]):
    """Abstract factory for creating envelope tracker instances."""

    @classmethod
    async def create_delivery_tracker(
        cls,
        cfg: Optional[C | dict[str, Any]] = None,
        storage_provider: Optional[StorageProvider] = None,
        tracker_store: Optional[KeyValueStore] = None,
        event_handler: Optional[DeliveryTrackerEventHandler] = None,
        retry_handler: Optional[RetryEventHandler] = None,
        **kwargs,
    ) -> Optional[DeliveryTracker]:
        """Create an envelope tracker instance based on the provided configuration."""

        return await create_resource(
            DeliveryTrackerFactory,
            cfg,
            storage_provider=storage_provider,
            kv_store=tracker_store,
            event_handler=event_handler,
            retry_handler=retry_handler,
            **kwargs,
        )



"""
Abstract factory base class for creating authorization policies.

Implementations of this factory create specific types of authorization
policies (e.g., expression-based, rule-based, etc.).
"""

from __future__ import annotations

from typing import Any, Optional, TypeVar

from pydantic import ConfigDict

from naylence.fame.factory import (
    ResourceConfig,
    ResourceFactory,
    create_default_resource,
    create_resource,
)
from naylence.fame.security.auth.policy.authorization_policy import (
    AuthorizationPolicy,
)

# Base type identifier for authorization policy factories
AUTHORIZATION_POLICY_FACTORY_BASE_TYPE = "AuthorizationPolicyFactory"


class AuthorizationPolicyConfig(ResourceConfig):
    """Configuration for creating an authorization policy."""

    model_config = ConfigDict(extra="allow")

    type: str


C = TypeVar("C", bound=AuthorizationPolicyConfig)


class AuthorizationPolicyFactory(ResourceFactory[AuthorizationPolicy, C]):
    """
    Abstract factory base class for creating authorization policies.

    Implementations of this factory create specific types of authorization
    policies (e.g., expression-based, rule-based, etc.).
    """

    @classmethod
    async def create_authorization_policy(
        cls,
        config: Optional[AuthorizationPolicyConfig | dict[str, Any]] = None,
    ) -> AuthorizationPolicy | None:
        """
        Static helper to create an authorization policy using the factory registry.

        Args:
            config: Configuration for the policy

        Returns:
            The created policy, or None if no factory matched

        Raises:
            ValueError: If config is provided but policy creation fails
        """
        if config:
            policy = await create_resource(AuthorizationPolicyFactory, config)

            if not policy:
                raise ValueError(
                    "Failed to create authorization policy from configuration"
                )

            return policy

        # Try to get default factory
        policy = await create_default_resource(AuthorizationPolicyFactory)

        return policy




"""
Default envelope tracker implementation that uses a pluggable KeyValueStore.

This implementation provides full tracking functionality while being agnostic
about the underlying storage mechanism - it can work with in-memory, persistent,
or any other KeyValueStore implementation.
"""

from __future__ import annotations

import asyncio
import os
import time
from typing import Any, AsyncIterator, Callable, Dict, Optional

from naylence.fame.core import (
    DeliveryAckFrame,
    DeliveryOriginType,
    FameDeliveryContext,
    FameEnvelope,
    FameResponseType,
    generate_id,
)
from naylence.fame.delivery.delivery_tracker import (
    DeliveryTracker,
    EnvelopeStatus,
    MailboxType,
    TrackedEnvelope,
)
from naylence.fame.delivery.retry_event_handler import RetryEventHandler
from naylence.fame.delivery.retry_policy import RetryPolicy
from naylence.fame.node.node_event_listener import NodeEventListener
from naylence.fame.node.node_like import NodeLike
from naylence.fame.storage.key_value_store import KeyValueStore
from naylence.fame.storage.storage_provider import StorageProvider
from naylence.fame.util import logging
from naylence.fame.util.formatter import AnsiColor, color, format_timestamp
from naylence.fame.util.task_spawner import TaskSpawner
from naylence.fame.util.util import pretty_model

logger = logging.getLogger(__name__)


_STREAM_END = object()

ENV_VAR_SHOW_ENVELOPES = "FAME_SHOW_ENVELOPES"

show_envelopes = bool(os.getenv(ENV_VAR_SHOW_ENVELOPES) == "true")


def _timestamp() -> str:
    return color(format_timestamp(), AnsiColor.GRAY)


class DefaultDeliveryTracker(NodeEventListener, DeliveryTracker, TaskSpawner):
    """
    Default envelope tracker implementation using a pluggable KeyValueStore.

    This implementation provides full tracking functionality including:
    - Ack/nack correlation with futures
    - Reply correlation via correlation IDs
    - Timeout and retry management
    - Event handler integration
    - Persistence via the provided KeyValueStore

    The storage mechanism is completely pluggable - use an in-memory store
    for testing/development or a persistent store for production.
    """

    def __init__(
        self,
        storage_provider: StorageProvider,
        # *,
        # retry_handler: Optional[RetryEventHandler] = None,
        futures_gc_grace_secs: int = 120,
        futures_sweep_interval_secs: int = 30,
    ) -> None:
        NodeEventListener.__init__(self)
        DeliveryTracker.__init__(self)
        TaskSpawner.__init__(self)
        self._storage_provider = storage_provider

        self._outbox: Optional[KeyValueStore[TrackedEnvelope]] = None

        self._inbox: Optional[KeyValueStore[TrackedEnvelope]] = None

        # corr_id -> envelope_id, use for replies only, not for ACKs
        self._correlation_to_envelope: dict[str, str] = {}

        self._timers: dict[str, asyncio.Task] = {}

        self._ack_futures: dict[str, asyncio.Future[FameEnvelope]] = {}
        self._reply_futures: dict[str, asyncio.Future[FameEnvelope]] = {}

        # GC bookkeeping for completed futures
        self._ack_done_since: dict[str, float] = {}
        self._reply_done_since: dict[str, float] = {}
        self._fut_gc_grace_secs = max(0, int(futures_gc_grace_secs))
        self._fut_sweep_interval_secs = max(1, int(futures_sweep_interval_secs))
        self._shutdown_event = asyncio.Event()

        self._stream_queues: dict[str, asyncio.Queue[Any]] = {}
        self._stream_done: dict[str, asyncio.Event] = {}

        # Dead-letter queue for inbound failures
        self._inbox_dlq: Optional[KeyValueStore[TrackedEnvelope]] = None

        self._lock = asyncio.Lock()
        self._node: NodeLike | None = None
        logger.debug("created_default_delivery_tracker")

    async def on_node_initialized(self, node: NodeLike) -> None:
        self._node = node
        self._outbox = await self._storage_provider.get_kv_store(
            model_cls=TrackedEnvelope,
            namespace="__delivery_outbox",
        )

        self._inbox = await self._storage_provider.get_kv_store(
            model_cls=TrackedEnvelope,
            namespace="__delivery_inbox",
        )

        # Initialize DLQ store for inbound envelopes
        self._inbox_dlq = await self._storage_provider.get_kv_store(
            model_cls=TrackedEnvelope,
            namespace="__delivery_inbox_dlq",
        )

    async def on_node_started(self, node: NodeLike) -> None:
        self._node = node
        # Start futures sweeper
        self.spawn(self._sweep_futures(), name="tracker-futures-sweeper")

    async def on_node_preparing_to_stop(self, node: NodeLike) -> None:
        """Wait for pending acknowledgments until their timeouts expire before node stops."""
        await self._wait_for_pending_acks()

    async def on_node_stopped(self, node: NodeLike) -> None:
        await self.cleanup()
        await self.shutdown_tasks()

    async def _wait_for_pending_acks(self) -> None:
        """Wait for pending acknowledgments until their timeouts expire."""
        logger.debug("tracker_node_preparing_to_stop_waiting_for_pending_acks")

        assert self._outbox

        # Get all pending envelopes that are expecting ACKs
        pending_acks = []
        async with self._lock:
            for envelope_id, future in list(self._ack_futures.items()):
                if not future.done():
                    pending_acks.append((envelope_id, future))

        if not pending_acks:
            logger.debug("tracker_no_pending_acks_to_wait_for")
            return

        logger.debug("tracker_waiting_for_pending_acks", count=len(pending_acks))

        # Wait for each ACK future with its individual timeout
        for envelope_id, future in pending_acks:
            try:
                # Get the tracked envelope to determine its timeout
                tracked = await self._outbox.get(envelope_id)
                if not tracked:
                    continue

                # Calculate remaining timeout
                now_ms = int(time.time() * 1000)
                remaining_ms = max(0, tracked.overall_timeout_at_ms - now_ms)

                if remaining_ms > 0:
                    timeout_seconds = remaining_ms / 1000.0
                    logger.debug(
                        "tracker_waiting_for_ack", envelope_id=envelope_id, timeout_seconds=timeout_seconds
                    )

                    try:
                        await asyncio.wait_for(future, timeout=timeout_seconds)
                        logger.debug("tracker_received_ack", envelope_id=envelope_id)
                    except asyncio.TimeoutError:
                        logger.debug("tracker_ack_timeout_expired", envelope_id=envelope_id)
                    except Exception as e:
                        logger.debug("tracker_ack_wait_error", envelope_id=envelope_id, error=str(e))
                    else:
                        await self._outbox.delete(envelope_id)
                else:
                    logger.debug("tracker_ack_already_expired", envelope_id=envelope_id)

            except Exception as e:
                logger.error("tracker_error_waiting_for_ack", envelope_id=envelope_id, error=str(e))

        logger.debug("tracker_finished_waiting_for_pending_acks")

    async def on_forward_upstream_complete(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        result: Optional[Any] = None,
        error: Optional[Exception] = None,
        context: FameDeliveryContext | None = None,
    ) -> Optional[FameEnvelope]:
        if show_envelopes:
            print(
                f"\n{_timestamp()} - {color('Forwarded envelope to upstream', AnsiColor.BLUE)} 🚀\n{
                    pretty_model(envelope)
                }"
            )
        return envelope

    async def on_forward_to_route_complete(
        self,
        node: NodeLike,
        next_segment: str,
        envelope: FameEnvelope,
        result: Optional[Any] = None,
        error: Optional[Exception] = None,
        context: FameDeliveryContext | None = None,
    ) -> Optional[FameEnvelope]:
        if show_envelopes:
            print(
                f"\n{_timestamp()} - {
                    color('Forwarded envelope to route "' + next_segment + '"', AnsiColor.BLUE)
                } 🚀\n{pretty_model(envelope)}"
            )
        return envelope

    async def on_forward_to_peer_complete(
        self,
        node: NodeLike,
        peer_segment: str,
        envelope: FameEnvelope,
        result: Optional[Any] = None,
        error: Optional[Exception] = None,
        context: FameDeliveryContext | None = None,
    ) -> Optional[FameEnvelope]:
        if show_envelopes:
            print(
                f"\n{_timestamp()} - '{
                    color('Forwarded envelope to peer "' + peer_segment + '"', AnsiColor.BLUE)
                }' 🚀\n{pretty_model(envelope)}"
            )
        return envelope

    async def on_heartbeat_sent(self, envelope: FameEnvelope) -> None:
        if show_envelopes:
            print(
                f"\n{_timestamp()} - {color('Sent envelope', AnsiColor.BLUE)} 🚀\n{pretty_model(envelope)}"
            )

    async def on_envelope_delivered(
        self, inbox_name: str, envelope: FameEnvelope, context: Optional[FameDeliveryContext] = None
    ) -> Optional[TrackedEnvelope]:
        logger.debug(
            "envelope_delivered",
            envp_id=envelope.id,
            corr_id=envelope.corr_id,
            rtype=FameResponseType(envelope.rtype) if envelope.rtype else FameResponseType.NONE,
            frame_type=type(envelope.frame).__name__,
        )

        assert self._outbox

        if not envelope.corr_id:
            logger.debug("envelope_delivered_no_corr_id", envelope_id=envelope.id)  # type: ignore
            return None

        if isinstance(envelope.frame, DeliveryAckFrame):
            # Ack handling
            if not envelope.frame.ref_id:
                logger.debug("envelope_delivered_no_ref_id", envelope_id=envelope.id)
                return

            if envelope.frame.ok:
                await self.on_ack(envelope, context)
            else:
                await self.on_nack(envelope, context)

        elif envelope.corr_id:
            return await self.on_correlated_message(inbox_name, envelope, context)

        return None

    async def on_envelope_handled(
        self, envelope: TrackedEnvelope, context: Optional[FameDeliveryContext] = None
    ) -> None:
        assert self._inbox
        # Set status to HANDLED for consistency
        envelope.status = EnvelopeStatus.HANDLED
        # Delete the envelope from inbox to prevent growth
        # Note: This is the new behavior to prevent inbox from growing indefinitely
        await self._inbox.delete(envelope.original_envelope.id)

    async def on_envelope_handle_failed(
        self,
        inbox_name: str,
        envelope: TrackedEnvelope,
        context: Optional[FameDeliveryContext] = None,
        error: Optional[Exception] = None,
        is_final_failure: bool = False,
    ) -> None:
        """
        Handle the case where envelope handling failed.
        This is called on every retry iteration, not just the final failure.
        Only sets FAILED_TO_HANDLE status if is_final_failure is True.
        """
        assert self._inbox

        # Add failure metadata without changing status (unless final failure)
        if error:
            envelope.meta[f"failure_attempt_{envelope.attempt}_reason"] = str(error)
            envelope.meta[f"failure_attempt_{envelope.attempt}_type"] = type(error).__name__
            envelope.meta["last_failure_reason"] = str(error)
            envelope.meta["last_failure_type"] = type(error).__name__

        if is_final_failure:
            envelope.status = EnvelopeStatus.FAILED_TO_HANDLE
            logger.error(
                "envelope_handle_failed_final",
                inbox_name=inbox_name,
                envp_id=envelope.original_envelope.id,
                error=str(error) if error else "unknown",
                status=envelope.status.name,
                total_attempts=envelope.attempt,
            )
            # Move to DLQ and remove from inbox
            await self.add_to_inbox_dlq(envelope, reason=str(error) if error else None)
            # Ensure it is no longer present in the active inbox
            await self._inbox.delete(envelope.original_envelope.id)
            return
        else:
            logger.warning(
                "envelope_handle_failed_retry",
                inbox_name=inbox_name,
                envp_id=envelope.original_envelope.id,
                error=str(error) if error else "unknown",
                attempt=envelope.attempt,
            )
            # Persist intermediate failure state so recovery can resume correctly
            await self._inbox.set(envelope.original_envelope.id, envelope)

    async def update_tracked_envelope(self, envelope: TrackedEnvelope) -> None:
        """Update a tracked envelope in persistent storage."""
        if envelope.mailbox_type == MailboxType.INBOX:
            assert self._inbox
            await self._inbox.update(envelope.original_envelope.id, envelope)
        elif envelope.mailbox_type == MailboxType.OUTBOX:
            raise RuntimeError(
                f"Updating tracked envelopes of mailbox type {MailboxType.OUTBOX} is not supported"
            )
        else:
            # Fallback for backwards compatibility - assume it's in inbox if mailbox_type is None
            assert self._inbox
            await self._inbox.update(envelope.original_envelope.id, envelope)

    async def _send_ack(self, envelope: FameEnvelope) -> None:
        assert self._node is not None

        if envelope.reply_to is None:
            logger.error("cannot_send_ack_no_reply_to", envp_id=envelope.id)
            return

        if envelope.corr_id is None:
            logger.error("cannot_send_ack_no_corr_id", envp_id=envelope.id)
            return

        logger.debug(
            "tracker_sending_ack", ref_id=envelope.id, to=envelope.reply_to, corr_id=envelope.corr_id
        )

        ack_env = self._node.envelope_factory.create_envelope(
            to=envelope.reply_to,
            frame=DeliveryAckFrame(ok=True, ref_id=envelope.id),
            corr_id=envelope.corr_id,
            trace_id=envelope.trace_id,
        )

        # Uncomment to simulate delayed ACK sending
        # async def delayed_send():
        #     assert self._node
        #     await asyncio.sleep(0.5)
        #     await self._node.send(ack_env)
        # asyncio.create_task(delayed_send())

        await self._node.send(ack_env)

    async def track(
        self,
        envelope: FameEnvelope,
        *,
        timeout_ms: int,
        expected_response_type: FameResponseType,
        retry_policy: Optional[RetryPolicy] = None,
        retry_handler: Optional[RetryEventHandler] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> Optional[TrackedEnvelope]:
        assert self._outbox

        now_ms = int(time.time() * 1000)

        if envelope.corr_id:
            corr_id = envelope.corr_id
        else:
            corr_id = envelope.corr_id = generate_id()

        async with self._lock:
            if envelope.id in self._ack_futures:
                logger.debug("tracker_envelope_already_tracked", envp_id=envelope.id)
                return None

            existing_env_id = self._correlation_to_envelope.get(envelope.corr_id)

            if expected_response_type & (FameResponseType.REPLY | FameResponseType.STREAM):
                if existing_env_id:
                    logger.debug(
                        "envelope_already_tracked_for_replies",
                        envp_id=envelope.id,
                        corr_id=corr_id,
                        expected_response_type=expected_response_type.name,
                    )
                    return None

                self._correlation_to_envelope[envelope.corr_id] = envelope.id

            # Create ack future if needed
            if expected_response_type & FameResponseType.ACK:
                self._ack_futures[envelope.id] = asyncio.get_running_loop().create_future()

            # Create reply future if needed
            if expected_response_type & FameResponseType.REPLY:
                self._reply_futures[envelope.id] = asyncio.get_running_loop().create_future()

            # Create stream end future if needed
            if expected_response_type & FameResponseType.STREAM:
                self._stream_queues[envelope.id] = asyncio.Queue()
                self._stream_done[envelope.id] = asyncio.Event()

        # Overall timeout is a hard cap; do not extend beyond timeout_ms
        overall_timeout_ms = timeout_ms

        # Determine the first timer checkpoint:
        # - With a retry policy, schedule the first retry callback at its first delay
        # - Without a retry policy, the next checkpoint is the overall timeout
        if retry_policy and retry_policy.max_retries > 0:
            try:
                first_delay_ms = max(0, int(retry_policy.next_delay_ms(1)))
            except Exception:
                first_delay_ms = 0
            first_checkpoint_ms = min(overall_timeout_ms, first_delay_ms)
        else:
            first_checkpoint_ms = overall_timeout_ms

        tracked = TrackedEnvelope(
            timeout_at_ms=now_ms + first_checkpoint_ms,
            overall_timeout_at_ms=now_ms + overall_timeout_ms,
            expected_response_type=expected_response_type,
            created_at_ms=now_ms,
            meta=meta or {},
            mailbox_type=MailboxType.OUTBOX,
            original_envelope=envelope,  # Store the envelope directly
        )

        # Persist to storage
        await self._outbox.set(envelope.id, tracked)

        # Schedule timeout/retry timer
        await self._schedule_timer(tracked, retry_policy, retry_handler=retry_handler)

        logger.debug(
            "tracker_registered_envelope",
            envp_id=envelope.id,
            corr_id=envelope.corr_id,
            expected_response=expected_response_type.name,
            target=str(envelope.to) if envelope.to else None,
            timeout_ms=timeout_ms,
        )
        return tracked

    async def await_ack(self, envelope_id: str, *, timeout_ms: Optional[int] = None) -> FameEnvelope:
        async with self._lock:
            future = self._ack_futures.get(envelope_id, None)

        if not future:
            raise RuntimeError(f"No ack expected for envelope {envelope_id}")

        return await self._await_envelope_future(
            envelope_id, FameResponseType.ACK, future, timeout_ms=timeout_ms
        )

    async def await_reply(self, envelope_id: str, *, timeout_ms: Optional[int] = None) -> Any:
        async with self._lock:
            future = self._reply_futures.get(envelope_id)

        if not future:
            raise RuntimeError(f"No reply expected for envelope {envelope_id}")

        return await self._await_envelope_future(
            envelope_id, FameResponseType.REPLY | FameResponseType.STREAM, future, timeout_ms=timeout_ms
        )

    async def _await_envelope_future(
        self,
        envelope_id: str,
        response_type: Any,
        future: asyncio.Future[FameEnvelope],
        timeout_ms: Optional[int] = None,
    ) -> FameEnvelope:
        assert self._outbox
        # Use provided timeout or calculate from tracked envelope
        if timeout_ms is None:
            # Try to get the envelope's configured timeout
            tracked = await self._outbox.get(envelope_id)
            if tracked:
                now_ms = int(time.time() * 1000)
                remaining_ms = max(0, tracked.overall_timeout_at_ms - now_ms)
                timeout_seconds = remaining_ms / 1000.0 if remaining_ms > 0 else None
            else:
                timeout_seconds = None
        else:
            timeout_seconds = timeout_ms / 1000.0

        result: FameEnvelope
        start_time = None
        try:
            if timeout_seconds is not None:
                result = await asyncio.wait_for(future, timeout=timeout_seconds)
            else:
                logger.debug("await_envelope_no_timeout_wait", envelope_id=envelope_id)
                result = await future
        except asyncio.TimeoutError as e:
            end_time = time.time()
            elapsed = end_time - start_time if start_time is not None else 0

            logger.error(
                "await_envelope_timeout_error",
                envelope_id=envelope_id,
                timeout_seconds=timeout_seconds,
                elapsed_seconds=elapsed,
                future_done=future.done(),
                future_cancelled=future.cancelled(),
                future_exception=future.exception() if future.done() and not future.cancelled() else None,
                error=str(e),
            )
            raise asyncio.TimeoutError(
                f"Timeout waiting for response_type({response_type}) for envelope {envelope_id}"
            )

        return result

    async def on_ack(self, envelope: FameEnvelope, context: Optional[FameDeliveryContext] = None) -> None:
        assert isinstance(envelope.frame, DeliveryAckFrame), "Ack must be from a DeliveryAckFrame"
        assert envelope.corr_id, "Reply envelope must have a correlation ID"
        assert envelope.frame.ref_id, "Ack frame must have a reference ID"
        assert self._outbox

        logger.debug(
            "tracker_on_ack", envp_id=envelope.id, corr_id=envelope.corr_id, ref_id=envelope.frame.ref_id
        )

        tracked_envelope = await self._outbox.get(envelope.frame.ref_id)

        if not tracked_envelope:
            logger.debug("tracker_ack_for_unknown_envelope", envp_id=envelope.id)
            return

        if tracked_envelope.original_envelope.corr_id != envelope.corr_id:
            logger.debug(
                "tracker_ack_corr_id_mismatch",
                envp_id=envelope.id,
                expected_corr_id=tracked_envelope.original_envelope.corr_id,
                actual_corr_id=envelope.corr_id,
            )
            return

        if tracked_envelope.original_envelope.id == envelope.id:
            # Received the original envelope instead of an ack, happens in local-to-local calls
            return

        # Update status
        tracked_envelope.status = (
            EnvelopeStatus.ACKED
            if not (tracked_envelope.expected_response_type & FameResponseType.STREAM)
            else tracked_envelope.status
        )
        await self._outbox.set(tracked_envelope.original_envelope.id, tracked_envelope)

        # Resolve ack future (idempotent) and mark for GC
        async with self._lock:
            future = self._ack_futures.get(tracked_envelope.original_envelope.id, None)
        if future and not future.done():
            future.set_result(envelope)
        await self._mark_done_since(
            self._ack_futures, tracked_envelope.original_envelope.id, self._ack_done_since
        )

        # Cancel timer
        await self._clear_timer(tracked_envelope.original_envelope.id)

        # Notify event handler
        for event_handler in self._event_handlers:
            await event_handler.on_envelope_acked(tracked_envelope)

        logger.debug("tracker_envelope_acked", envp_id=tracked_envelope.original_envelope.id)

    async def on_nack(self, envelope: FameEnvelope, context: Optional[FameDeliveryContext] = None) -> None:
        assert isinstance(envelope.frame, DeliveryAckFrame), "Nack must be from a DeliveryAckFrame"
        assert envelope.corr_id, "Reply envelope must have a correlation ID"
        assert envelope.frame.ref_id, "Ack frame must have a reference ID"
        assert self._outbox

        tracked_envelope = await self._outbox.get(envelope.frame.ref_id)

        if not tracked_envelope:
            logger.debug("tracker_nack_for_unknown_envelope", envp_id=envelope.id)
            return

        if tracked_envelope.original_envelope.corr_id != envelope.corr_id:
            logger.debug(
                "tracker_nack_corr_id_mismatch",
                envp_id=envelope.id,
                expected_corr_id=tracked_envelope.original_envelope.corr_id,
                actual_corr_id=envelope.corr_id,
            )
            return

        # Update status and metadata
        tracked_envelope.status = EnvelopeStatus.NACKED
        if envelope.frame.reason:
            tracked_envelope.meta["nack_reason"] = envelope.frame.reason

        await self._outbox.set(tracked_envelope.original_envelope.id, tracked_envelope)

        # Resolve ack future with error (idempotent) and mark for GC
        async with self._lock:
            future = self._ack_futures.get(tracked_envelope.original_envelope.id, None)
        if future and not future.done():
            future.set_exception(RuntimeError(f"Envelope nacked: {envelope.frame.reason or 'unknown'}"))
        await self._mark_done_since(
            self._ack_futures, tracked_envelope.original_envelope.id, self._ack_done_since
        )

        stream_queue = self._stream_queues.get(tracked_envelope.original_envelope.id)
        if stream_queue:
            await stream_queue.put(envelope)
            await stream_queue.put(_STREAM_END)
            ev = self._stream_done.get(tracked_envelope.original_envelope.id)
            if ev:
                ev.set()

        # Cancel timer
        await self._clear_timer(tracked_envelope.original_envelope.id)

        # Notify event handler
        for event_handler in self._event_handlers:
            await event_handler.on_envelope_nacked(tracked_envelope, envelope.frame.reason)

        logger.debug(
            "tracker_envelope_nacked",
            envp_id=tracked_envelope.original_envelope.id,
            reason=envelope.frame.reason,
        )

    async def on_correlated_message(
        self, inbox_name: str, envelope: FameEnvelope, context: Optional[FameDeliveryContext] = None
    ) -> Optional[TrackedEnvelope]:
        assert envelope.corr_id, "Envelope must have a correlation ID"
        assert self._outbox

        async with self._lock:
            tracked_envelope_id = self._correlation_to_envelope.get(envelope.corr_id)

        if (
            tracked_envelope_id is not None
            and (tracked_envelope := await self._outbox.get(tracked_envelope_id))
            and tracked_envelope.original_envelope.id != envelope.id
        ):
            tracked_envelope = await self.on_reply(envelope, tracked_envelope, context)
        else:
            assert self._inbox

            tracked_envelope = await self._inbox.get(envelope.id)
            if tracked_envelope is not None:
                if tracked_envelope.status != EnvelopeStatus.HANDLED:
                    tracked_envelope.status = EnvelopeStatus.RECEIVED
                    await self._inbox.set(envelope.id, tracked_envelope)
                else:
                    logger.debug(
                        "tracker_duplicate_envelope_already_handled",
                        envp_id=envelope.id,
                        status=tracked_envelope.status.name,
                    )
            else:
                tracked_envelope = TrackedEnvelope(
                    timeout_at_ms=0,
                    overall_timeout_at_ms=0,
                    expected_response_type=envelope.rtype or FameResponseType.NONE,
                    created_at_ms=int(time.time() * 1000),
                    status=EnvelopeStatus.RECEIVED,
                    mailbox_type=MailboxType.INBOX,
                    original_envelope=envelope,
                    service_name=inbox_name,
                )

                await self._inbox.set(envelope.id, tracked_envelope)

        # When reply itself requires an ack, send it
        if envelope.rtype and bool(envelope.rtype & FameResponseType.ACK):
            await self._send_ack(envelope)

        return tracked_envelope

    async def on_reply(
        self,
        envelope: FameEnvelope,
        tracked_envelope: TrackedEnvelope,
        context: Optional[FameDeliveryContext] = None,
    ) -> TrackedEnvelope:
        assert self._node is not None
        assert envelope.corr_id, "Reply envelope must have a correlation ID"
        assert self._outbox
        assert tracked_envelope.original_envelope.id != envelope.id

        if tracked_envelope.expected_response_type & FameResponseType.STREAM:
            # Treat as stream item for metrics only; upstream handles delivery
            await self.on_stream_item(tracked_envelope.original_envelope.id, envelope)
            return tracked_envelope

        # Update status
        tracked_envelope.status = EnvelopeStatus.RESPONDED
        await self._outbox.set(tracked_envelope.original_envelope.id, tracked_envelope)

        # Cancel timer
        await self._clear_timer(tracked_envelope.original_envelope.id)

        # Resolve reply/ack futures (idempotent) and mark for GC
        async with self._lock:
            reply_future = self._reply_futures.get(tracked_envelope.original_envelope.id, None)
            ack_future = self._ack_futures.get(tracked_envelope.original_envelope.id, None)

        if reply_future and not reply_future.done():
            reply_future.set_result(envelope)
        await self._mark_done_since(
            self._reply_futures, tracked_envelope.original_envelope.id, self._reply_done_since
        )

        if ack_future and not ack_future.done():
            ack_env = self._node.envelope_factory.create_envelope(
                to=envelope.reply_to,
                frame=DeliveryAckFrame(
                    ok=True,
                    ref_id=tracked_envelope.original_envelope.id,
                    reason="Auto-ack for reply",
                ),
                corr_id=envelope.corr_id,
                trace_id=envelope.trace_id,
            )
            ack_future.set_result(ack_env)
        await self._mark_done_since(
            self._ack_futures, tracked_envelope.original_envelope.id, self._ack_done_since
        )

        # Notify event handler
        for event_handler in self._event_handlers:
            await event_handler.on_envelope_replied(tracked_envelope, envelope)

        logger.debug(
            "tracked_envelope_replied",
            envp_id=tracked_envelope.original_envelope.id,
            corr_id=envelope.corr_id,
        )
        return tracked_envelope

    async def iter_stream(
        self, envelope_id: str, *, timeout_ms: Optional[int] = None
    ) -> AsyncIterator[Any]:
        stream_queue = self._stream_queues.get(envelope_id)
        done = self._stream_done.get(envelope_id)
        if not stream_queue or not done:
            # Not a stream-tracked envelope
            return
            # yield  # make function an iterator

        per_get_timeout = (timeout_ms / 1000.0) if timeout_ms else None
        while True:
            try:
                item = (
                    await asyncio.wait_for(stream_queue.get(), timeout=per_get_timeout)
                    if per_get_timeout
                    else await stream_queue.get()
                )
            except asyncio.TimeoutError as e:
                raise asyncio.TimeoutError(f"stream timeout waiting for next item: {e}")
            if item is _STREAM_END:
                break
            if isinstance(item, Exception):
                raise item
            yield item

    async def on_stream_item(self, envelope_id: str, reponse_env: FameEnvelope) -> None:
        q = self._stream_queues.get(envelope_id)
        if not q:
            return
        logger.debug("tracker_stream_item", envp_id=envelope_id, response_envp_id=reponse_env.id)
        await q.put(reponse_env)

    async def on_stream_end(self, envelope_id: str) -> None:
        assert self._outbox
        entry = await self._outbox.get(envelope_id)
        if entry:
            entry.status = EnvelopeStatus.RESPONDED
            await self._outbox.set(envelope_id, entry)
        q = self._stream_queues.get(envelope_id)
        if q:
            await q.put(_STREAM_END)
        ev = self._stream_done.get(envelope_id)
        if ev:
            ev.set()

    async def get_tracked_envelope(self, envelope_id: str) -> Optional[TrackedEnvelope]:
        assert self._outbox
        entry = await self._outbox.get(envelope_id)
        return entry

    async def list_pending(self) -> list[TrackedEnvelope]:
        assert self._outbox
        all_entries = await self._outbox.list()
        pending = [entry for entry in all_entries.values() if entry.status == EnvelopeStatus.PENDING]
        return pending

    async def list_inbound(
        self, filter: Optional[Callable[[TrackedEnvelope], bool]] = None
    ) -> list[TrackedEnvelope]:
        """List inbound envelopes that are in RECEIVED or FAILED_TO_HANDLE status."""
        if not self._inbox:
            # Return empty list if inbox is not initialized yet
            return []
        all_entries = await self._inbox.list()
        failed_inbound = [entry for entry in all_entries.values() if not filter or filter(entry)]
        return failed_inbound

    # Inbox DLQ API
    async def add_to_inbox_dlq(
        self, tracked_envelope: TrackedEnvelope, reason: Optional[str] = None
    ) -> None:
        """
        Put an inbound tracked envelope into the DLQ store and annotate with metadata.
        Keeps the envelope payload for later inspection or manual requeue.
        """
        if not self._inbox_dlq:
            logger.error("dlq_not_initialized", envp_id=tracked_envelope.original_envelope.id)
            return

        # Mark DLQ metadata (do not introduce a new status to avoid breaking consumers)
        tracked_envelope.meta["dlq"] = True
        if reason:
            tracked_envelope.meta["dlq_reason"] = reason
        tracked_envelope.meta["dead_lettered_at_ms"] = int(time.time() * 1000)

        await self._inbox_dlq.set(tracked_envelope.original_envelope.id, tracked_envelope)
        logger.warning(
            "envelope_moved_to_dlq",
            envp_id=tracked_envelope.original_envelope.id,
            service_name=tracked_envelope.service_name,
        )

    async def get_from_inbox_dlq(self, envelope_id: str) -> Optional[TrackedEnvelope]:
        """Get a specific envelope from the inbox DLQ by envelope ID."""
        if not self._inbox_dlq:
            return None
        return await self._inbox_dlq.get(envelope_id)

    async def list_inbox_dlq(self) -> list[TrackedEnvelope]:
        """List all envelopes currently in the inbox DLQ."""
        if not self._inbox_dlq:
            return []
        items = await self._inbox_dlq.list()
        return list(items.values())

    async def purge_inbox_dlq(self, predicate: Optional[Callable[[TrackedEnvelope], bool]] = None) -> int:
        """
        Delete inbox DLQ entries. If a predicate is provided, only delete matching items.
        Returns the number of deleted entries.
        """
        if not self._inbox_dlq:
            return 0
        items = await self._inbox_dlq.list()
        to_delete = [k for k, v in items.items() if (predicate(v) if predicate else True)]
        for key in to_delete:
            await self._inbox_dlq.delete(key)
        if to_delete:
            logger.debug("dlq_purged", count=len(to_delete))
        return len(to_delete)

    async def cleanup(self) -> None:
        # Signal shutdown to wake up the sweeper
        self._shutdown_event.set()

        # Cancel all timers
        async with self._lock:
            timers = list(self._timers.values())
            self._timers.clear()

            # Cancel all ack futures
            for future in self._ack_futures.values():
                if not future.done():
                    future.cancel()
            self._ack_futures.clear()
            self._ack_done_since.clear()

            # Cancel all reply futures
            for future in self._reply_futures.values():
                if not future.done():
                    future.cancel()
            self._reply_futures.clear()
            self._reply_done_since.clear()

            for q in self._stream_queues.values():
                # signal end to any iterators
                await q.put(_STREAM_END)

            self._stream_queues.clear()

            for ev in self._stream_done.values():
                ev.set()

            self._stream_done.clear()

            self._correlation_to_envelope.clear()

        # Wait for timers to complete
        for timer in timers:
            timer.cancel()
            try:
                await timer
            except asyncio.CancelledError:
                pass

        logger.debug("tracker_cleanup_completed")

    async def recover_pending(self) -> None:
        """Recover pending envelopes and reschedule timers."""
        pending = await self.list_pending()
        logger.debug("tracker_recovering_pending", count=len(pending))

        async with self._lock:
            # Rebuild correlation mapping
            for tracked in pending:
                # Recreate ack future if needed
                if tracked.expected_response_type & FameResponseType.ACK:
                    self._ack_futures[tracked.original_envelope.id] = (
                        asyncio.get_running_loop().create_future()
                    )

                # Recreate reply future if needed
                if tracked.expected_response_type & FameResponseType.REPLY:
                    if tracked.original_envelope.corr_id:
                        self._correlation_to_envelope[tracked.original_envelope.corr_id] = (
                            tracked.original_envelope.id
                        )
                    self._reply_futures[tracked.original_envelope.id] = (
                        asyncio.get_running_loop().create_future()
                    )

                if tracked.expected_response_type & FameResponseType.STREAM:
                    if tracked.original_envelope.corr_id:
                        self._correlation_to_envelope[tracked.original_envelope.corr_id] = (
                            tracked.original_envelope.id
                        )
                    self._stream_queues[tracked.original_envelope.id] = asyncio.Queue()
                    self._stream_done[tracked.original_envelope.id] = asyncio.Event()

        # Reschedule timers (no retry policy on recovery)
        for tracked in pending:
            await self._schedule_timer(tracked, retry_policy=None, retry_handler=None)

        logger.debug("tracker_recovery_completed", count=len(pending))

    async def _schedule_timer(
        self,
        tracked: TrackedEnvelope,
        retry_policy: Optional[RetryPolicy],
        retry_handler: Optional[RetryEventHandler] = None,
    ) -> None:
        """Schedule a timeout/retry timer for an envelope."""
        assert self._outbox

        async with self._lock:
            # Cancel existing timer
            existing_timer = self._timers.get(tracked.original_envelope.id)
            if existing_timer:
                existing_timer.cancel()

            async def _timer():
                assert self._outbox
                assert self._node
                try:
                    now_ms = int(time.time() * 1000)

                    # Determine what happens first: retry or overall timeout
                    next_retry_at_ms = tracked.timeout_at_ms
                    overall_timeout_at_ms = tracked.overall_timeout_at_ms

                    if next_retry_at_ms <= overall_timeout_at_ms:
                        # Wait for retry/first timeout
                        delay_ms = max(0, next_retry_at_ms - now_ms)
                    else:
                        # Wait for overall timeout
                        delay_ms = max(0, overall_timeout_at_ms - now_ms)

                    if delay_ms > 0:
                        await asyncio.sleep(delay_ms / 1000.0)

                    # Check current status
                    entry = await self._outbox.get(tracked.original_envelope.id)
                    if not entry or entry.status != EnvelopeStatus.PENDING:
                        return

                    current_tracked = entry
                    now_ms = int(time.time() * 1000)

                    # Check if we've exceeded the overall timeout
                    if now_ms >= current_tracked.overall_timeout_at_ms:
                        # Overall timeout - no more retries
                        current_tracked.status = EnvelopeStatus.TIMED_OUT
                        await self._outbox.set(tracked.original_envelope.id, current_tracked)

                        # Cancel futures by completing with TimeoutError (idempotent) and mark for GC
                        async with self._lock:
                            future = self._ack_futures.get(tracked.original_envelope.id, None)
                        if future and not future.done():
                            future.set_exception(asyncio.TimeoutError())
                        await self._mark_done_since(
                            self._ack_futures, tracked.original_envelope.id, self._ack_done_since
                        )

                        async with self._lock:
                            reply_future = self._reply_futures.get(tracked.original_envelope.id, None)
                        if reply_future and not reply_future.done():
                            reply_future.set_exception(asyncio.TimeoutError())
                        await self._mark_done_since(
                            self._reply_futures, tracked.original_envelope.id, self._reply_done_since
                        )

                        # Notify event handlers
                        for event_handler in self._event_handlers:
                            await event_handler.on_envelope_timeout(current_tracked)

                        logger.debug("tracker_envelope_timed_out", envp_id=tracked.original_envelope.id)

                    elif retry_policy and current_tracked.attempt < retry_policy.max_retries:
                        # Schedule retry
                        current_tracked.attempt += 1
                        next_delay_ms = retry_policy.next_delay_ms(current_tracked.attempt)

                        # Set next retry time, but don't exceed overall timeout
                        next_retry_time = now_ms + next_delay_ms
                        if next_retry_time <= current_tracked.overall_timeout_at_ms:
                            current_tracked.timeout_at_ms = next_retry_time
                        else:
                            # Next retry would exceed overall timeout, so schedule final timeout instead
                            current_tracked.timeout_at_ms = current_tracked.overall_timeout_at_ms

                        # Update storage
                        await self._outbox.set(tracked.original_envelope.id, current_tracked)

                        # Notify retry handler
                        if retry_handler and current_tracked.original_envelope:
                            await retry_handler.on_retry_needed(
                                current_tracked.original_envelope,
                                current_tracked.attempt,
                                next_delay_ms,
                                context=FameDeliveryContext(
                                    from_system_id=self._node.id, origin_type=DeliveryOriginType.LOCAL
                                ),
                            )

                        # Reschedule timer
                        await self._schedule_timer(current_tracked, retry_policy, retry_handler)

                        logger.debug(
                            "envelope_delivery_retry_scheduled",
                            envp_id=tracked.original_envelope.id,
                            attempt=current_tracked.attempt,
                            max_retries=retry_policy.max_retries,
                            next_delay_ms=next_delay_ms,
                        )
                    else:
                        # No more retries available; keep waiting until the overall timeout cap
                        if now_ms < current_tracked.overall_timeout_at_ms:
                            current_tracked.timeout_at_ms = current_tracked.overall_timeout_at_ms
                            await self._outbox.set(tracked.original_envelope.id, current_tracked)

                            # Reschedule only to the overall timeout; no further retries will be issued
                            await self._schedule_timer(current_tracked, retry_policy, retry_handler)

                            logger.debug(
                                "envelope_retries_exhausted_waiting_until_overall_timeout",
                                envp_id=tracked.original_envelope.id,
                                attempt=current_tracked.attempt,
                                overall_timeout_at_ms=current_tracked.overall_timeout_at_ms,
                            )
                            return

                        # Fallback: treat as timed out (should normally be handled earlier)
                        current_tracked.status = EnvelopeStatus.TIMED_OUT
                        await self._outbox.set(tracked.original_envelope.id, current_tracked)

                        # Cancel futures by completing with TimeoutError (idempotent) and mark for GC
                        async with self._lock:
                            future = self._ack_futures.get(tracked.original_envelope.id, None)
                        if future and not future.done():
                            future.set_exception(asyncio.TimeoutError())
                        await self._mark_done_since(
                            self._ack_futures, tracked.original_envelope.id, self._ack_done_since
                        )

                        async with self._lock:
                            reply_future = self._reply_futures.get(tracked.original_envelope.id, None)
                        if reply_future and not reply_future.done():
                            reply_future.set_exception(asyncio.TimeoutError())
                        await self._mark_done_since(
                            self._reply_futures, tracked.original_envelope.id, self._reply_done_since
                        )

                        # Notify event handlers
                        for event_handler in self._event_handlers:
                            await event_handler.on_envelope_timeout(current_tracked)

                        logger.debug("tracker_envelope_timed_out", envp_id=tracked.original_envelope.id)

                except asyncio.CancelledError:
                    pass
                except Exception as e:
                    logger.error("tracker_timer_error", envp_id=tracked.original_envelope.id, error=str(e))

            task = self.spawn(_timer(), name=f"tracker-{tracked.original_envelope.id}")
            self._timers[tracked.original_envelope.id] = task

    async def _clear_timer(self, envelope_id: str) -> None:
        """Cancel and remove a timer for an envelope."""
        async with self._lock:
            timer = self._timers.pop(envelope_id, None)

        if timer:
            try:
                timer.cancel()
            except Exception:
                # Ignore exceptions from cancel - timer might be in unexpected state
                pass
            try:
                await timer
            except asyncio.CancelledError:
                pass

    def _status_is_terminal(self, status: EnvelopeStatus) -> bool:
        return status in (
            EnvelopeStatus.ACKED,
            EnvelopeStatus.RESPONDED,
            EnvelopeStatus.TIMED_OUT,
            EnvelopeStatus.NACKED,
        )

    async def _mark_done_since(
        self,
        registry: dict[str, asyncio.Future[Any]],
        env_id: str,
        done_since_map: dict[str, float],
    ) -> None:
        # Only record first done timestamp
        async with self._lock:
            fut = registry.get(env_id)
            if fut and fut.done() and env_id not in done_since_map:
                done_since_map[env_id] = time.time()

    async def _sweep_futures(self) -> None:
        # Periodically sweep completed futures after grace period,
        # only when the tracked envelope is terminal or absent.
        while True:
            try:
                # Wait for either timeout or shutdown signal
                try:
                    await asyncio.wait_for(
                        self._shutdown_event.wait(), timeout=self._fut_sweep_interval_secs
                    )
                    # Shutdown event was set, exit
                    break
                except asyncio.TimeoutError:
                    # Normal timeout, continue with sweep
                    pass

                if not self._outbox:
                    continue

                now = time.time()
                # Snapshot candidates without holding the lock across awaits
                async with self._lock:
                    ack_candidates = [
                        env_id
                        for (env_id, since) in self._ack_done_since.items()
                        if now - since >= self._fut_gc_grace_secs
                    ]
                    reply_candidates = [
                        env_id
                        for (env_id, since) in self._reply_done_since.items()
                        if now - since >= self._fut_gc_grace_secs
                    ]

                # Check terminal state
                def _batch(items: list[str], size: int = 128):
                    for i in range(0, len(items), size):
                        yield items[i : i + size]

                to_remove_ack: list[str] = []
                to_remove_reply: list[str] = []

                # Evaluate ACK candidates
                for batch in _batch(ack_candidates):
                    for env_id in batch:
                        tracked = await self._outbox.get(env_id)
                        if tracked is None or self._status_is_terminal(tracked.status):
                            to_remove_ack.append(env_id)

                # Evaluate REPLY candidates
                for batch in _batch(reply_candidates):
                    for env_id in batch:
                        tracked = await self._outbox.get(env_id)
                        if tracked is None or self._status_is_terminal(tracked.status):
                            to_remove_reply.append(env_id)

                # Remove under lock
                if to_remove_ack or to_remove_reply:
                    async with self._lock:
                        for env_id in to_remove_ack:
                            self._ack_futures.pop(env_id, None)
                            self._ack_done_since.pop(env_id, None)
                        for env_id in to_remove_reply:
                            self._reply_futures.pop(env_id, None)
                            self._reply_done_since.pop(env_id, None)

                    if to_remove_ack or to_remove_reply:
                        logger.debug(
                            "tracker_swept_completed_futures",
                            ack_removed=len(to_remove_ack),
                            reply_removed=len(to_remove_reply),
                            grace_secs=self._fut_gc_grace_secs,
                        )
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error("tracker_sweeper_error", error=str(e))



from typing import Optional, Protocol, Sequence, runtime_checkable

from naylence.fame.core import FameEnvelope, Stickiness


@runtime_checkable
class LoadBalancerStickinessManager(Protocol):
    def negotiate(self, stickiness: Optional[Stickiness]) -> Optional[Stickiness]: ...

    def get_sticky_replica_segment(
        self, envelope: FameEnvelope, segments: Optional[Sequence[str]] = None
    ) -> Optional[str]: ...



from __future__ import annotations

import asyncio
from asyncio import Queue
from typing import Optional

from naylence.fame.core import (
    DEFAULT_POLLING_TIMEOUT_MS,
    Channel,
    FameEnvelope,
    ReadWriteChannel,
)


def in_memory_channel_factory(spec: dict):
    return InMemoryChannel(spec)


class InMemoryChannel(Channel):
    def __init__(self, spec: dict):
        self._spec = spec
        self._queue: Queue[FameEnvelope] = Queue()


class InMemoryReadWriteChannel(ReadWriteChannel):
    def __init__(self, queue: Optional[Queue[FameEnvelope]] = None):
        self._queue = queue if queue is not None else Queue()

    async def receive(self, timeout: int = DEFAULT_POLLING_TIMEOUT_MS) -> Optional[FameEnvelope]:
        try:
            return await asyncio.wait_for(self._queue.get(), timeout=timeout / 1000.0)
        except asyncio.TimeoutError:
            return None

    async def acknowledge(self, message_id: str) -> None:
        # No-op for now
        pass

    async def send(self, message: FameEnvelope) -> None:
        await self._queue.put(message)
        return None



"""
Default implementation of KeyManager.
"""

import asyncio
from pathlib import PurePosixPath
from typing import TYPE_CHECKING, Any, Iterable, Optional

from naylence.fame.core import (
    DeliveryOriginType,
    EnvelopeFactory,
    KeyAnnounceFrame,
    local_delivery_context,
)
from naylence.fame.security.keys.key_store import KeyStore
from naylence.fame.util.envelope_context import current_trace_id
from naylence.fame.util.logging import getLogger
from naylence.fame.util.util import secure_digest

from .key_manager import KeyManager

if TYPE_CHECKING:
    from naylence.fame.node.node_like import NodeLike
    from naylence.fame.node.routing_node_like import RoutingNodeLike

logger = getLogger(__name__)


class DefaultKeyManager(KeyManager):
    """Default implementation of KeyManager with certificate expiry management."""

    def __init__(
        self,
        *,
        key_store: KeyStore,
    ) -> None:
        KeyManager.__init__(self)
        self._key_store: KeyStore = key_store

        # Context - populated by on_node_started
        self._node: Optional[NodeLike] = None
        self._routing_node: Optional[RoutingNodeLike] = None

    async def on_node_started(self, node: "NodeLike") -> None:
        """Called when a node has been started and is ready for operation.

        This replaces the deprecated update_context() method and also starts
        the background certificate purging task.
        """
        self._node = node

        # Check if this is a routing node for additional capabilities
        try:
            from naylence.fame.node.routing_node_like import RoutingNodeLike

            if isinstance(node, RoutingNodeLike):
                self._routing_node = node
        except ImportError:
            # RoutingNodeLike not available, skip routing-specific features
            pass

        logger.debug(
            "key_manager_started",
            node_id=self._node_id,
            physical_path=self._physical_path,
            has_upstream=self._has_upstream,
        )

        # Start background certificate purging task

    async def on_node_stopped(self, node: "NodeLike") -> None:
        """Called when a node is being stopped and should clean up resources."""
        logger.debug("key_manager_stopped", node_id=self._node_id)

    @property
    def _has_upstream(self) -> bool:
        return self._node.has_parent if self._node else False

    @property
    def _physical_path(self) -> str:
        """Get the node's physical path."""
        return str(getattr(self._node, "physical_path", "/")) if self._node else "/"

    @property
    def _node_id(self) -> str:
        """Get the node's ID."""
        return str(getattr(self._node, "_id", "")) if self._node else ""

    @property
    def _node_sid(self) -> str:
        """Get the node's secure ID."""
        return str(getattr(self._node, "_sid", "")) if self._node else ""

    @property
    def _envelope_factory(self) -> Optional[EnvelopeFactory]:
        """Get the node's envelope factory."""
        return getattr(self._node, "_envelope_factory", None) if self._node else None

    async def get_key(self, kid: str) -> dict[str, Any]:
        return await self._key_store.get_key(kid)

    async def has_key(self, kid: str) -> bool:
        return await self._key_store.has_key(kid)

    async def add_keys(
        self,
        *,
        keys: list[dict],
        sid: Optional[str] = None,
        physical_path: str,
        system_id: str,
        origin: DeliveryOriginType,
        skip_sid_validation: bool = False,  # New parameter for correlation routing
    ):
        from naylence.fame.security.crypto.jwk_validation import (
            JWKValidationError,
            validate_jwk_complete,
        )

        # Pre-validate all keys before processing
        valid_keys = []
        rejected_count = 0  # Get trust store configuration if available

        for key in keys:
            try:
                # First, basic JWK validation
                validate_jwk_complete(key)

                valid_keys.append(key)

            except JWKValidationError as e:
                logger.warning(
                    "rejected_invalid_jwk_in_announce",
                    kid=key.get("kid", "unknown"),
                    from_system_id=system_id,
                    from_physical_path=physical_path,
                    error=str(e),
                )
                rejected_count += 1
        if not valid_keys:
            logger.warning(
                "no_valid_keys_in_announce",
                from_system_id=system_id,
                from_physical_path=physical_path,
                total_keys=len(keys),
                rejected_count=rejected_count,
            )
            return

        logger.debug(
            "adding_keys",
            key_ids=[key["kid"] for key in valid_keys],
            source_system_id=system_id,
            from_physical_path=physical_path,
            trace_id=current_trace_id(),
            origin=origin,
            valid_count=len(valid_keys),
            rejected_count=rejected_count,
        )

        if origin == DeliveryOriginType.LOCAL:
            await self._key_store.add_keys(valid_keys, physical_path=physical_path)
            return

        self_physical_path = self._physical_path

        # Accept key announcements signed by direct links only
        # Skip SID validation for correlation-routed KeyAnnounce messages
        if sid and not skip_sid_validation:
            key_path = None

            if origin == DeliveryOriginType.DOWNSTREAM:
                key_path = str(PurePosixPath(self_physical_path) / system_id)
            elif origin == DeliveryOriginType.UPSTREAM:
                key_path = str(PurePosixPath(self_physical_path).parent)
            elif origin == DeliveryOriginType.PEER:
                key_path = f"/{system_id}"

            assert key_path

            expected_sid = secure_digest(key_path)
            if sid != expected_sid:
                # TODO: security: disconnect the downstream that sent the announcement?
                raise ValueError(f"Invalid downstream sid: {sid}")

        # For downstream origins, ensure the physical path is downstream
        if origin == DeliveryOriginType.DOWNSTREAM:
            normalized_frame_path = physical_path.rstrip("/")
            expected_path_prefix = str(PurePosixPath(self_physical_path) / system_id / "")
            if not (normalized_frame_path + "/").startswith(expected_path_prefix):
                raise ValueError(
                    f"Frame physical path {normalized_frame_path} "
                    f"does not match expected prefix {expected_path_prefix}"
                )

        await self._key_store.add_keys(valid_keys, physical_path=physical_path)

        if origin == DeliveryOriginType.DOWNSTREAM:
            await self._announce_path_keys(valid_keys, physical_path, origin)
        else:
            logger.debug(
                "skip_announcing_keys_to_upstream",
                key_ids=[key["kid"] for key in keys if "kid" in key],
                from_physical_path=physical_path,
                has_upsteam=self._has_upstream,
            )

    async def _announce_path_keys(
        self,
        keys: list[dict[str, Any]],
        from_physical_path: str,
        origin: DeliveryOriginType,
    ):
        # Only announce if we have somewhere to announce to
        if not (
            self._has_upstream or (self._routing_node and hasattr(self._routing_node, "forward_to_peers"))
        ):
            logger.debug(
                "skip_announcing_keys_no_destination",
                key_ids=[key["kid"] for key in keys if "kid" in key],
                from_physical_path=from_physical_path,
                has_upstream=self._has_upstream,
                has_routing_node=self._routing_node is not None,
            )
            return

        logger.debug(
            "announcing_keys_to_upstream",
            key_ids=[key["kid"] for key in keys if "kid" in key],
            from_physical_path=from_physical_path,
        )

        envelope_factory = self._envelope_factory
        if not envelope_factory:
            raise RuntimeError("Envelope factory not available - key manager not properly initialized")

        env = envelope_factory.create_envelope(
            frame=KeyAnnounceFrame(keys=keys, physical_path=from_physical_path)
        )

        if self._has_upstream:
            node_id = self._node_id
            if not node_id:
                raise RuntimeError("Node ID not available - key manager not properly initialized")
            if not self._node:
                raise RuntimeError("Node not available - key manager not properly initialized")
            await self._node.forward_upstream(env, local_delivery_context(node_id))

        if self._routing_node and hasattr(self._routing_node, "forward_to_peers"):
            node_id = self._node_id
            if not node_id:
                raise RuntimeError("Node ID not available - key manager not properly initialized")
            await self._routing_node.forward_to_peers(env, None, None, local_delivery_context(node_id))

    async def announce_keys_to_upstream(self):
        if not self._has_upstream:
            return

        self_physical_path = self._physical_path

        logger.debug("reannouncing_keys_upstream")

        sem = asyncio.Semaphore(16)  # throttle burst to 16 in-flight

        async def _announce(path: str, keys_per_sid: list[dict[str, Any]]) -> None:
            if not path.startswith(self_physical_path):
                return
            async with sem:
                try:
                    await self._announce_path_keys(keys_per_sid, path, DeliveryOriginType.DOWNSTREAM)
                except Exception as e:
                    logger.error("announce_key_failed", error=e, exc_info=True)

        # TODO: get keys relevant to the origin
        await asyncio.gather(
            *(
                _announce(path, keys)
                for path, keys in (await self._key_store.get_keys_grouped_by_path()).items()
            )
        )

        logger.debug("reannounce_keys_upstream_completed")

    def _get_keys_to_announce_upstream(self) -> list[list[dict[str, Any]]]: ...

    def _should_announce_upsteam(self, key: dict[str, Any]) -> bool:
        return False

    async def handle_key_request(
        self,
        kid: str,
        from_seg: str,
        *,
        physical_path: str | None,
        origin: DeliveryOriginType,
        corr_id: Optional[str] = None,
        original_client_sid: Optional[str] = None,
    ) -> None:
        """
        Answer a KeyRequest with a signed KeyAnnounce.

        Parameters
        ----------
        kid : str
            The requested key-id.
        from_seg : str
            The segment we received the request *from* (so we can target the reply
            if it came from downstream).
        physical_path : str | None
            Optional path filter for bulk requests.
        origin : DeliveryOriginType
            UPSTREAM if the request travelled down from our parent,
            DOWNSTREAM if it came up from a child.
        corr_id : str | None
            Optional correlation ID for the request.
        original_client_sid : str | None
            Original client session ID for sticky routing.
        """

        logger.trace("handling_key_request", kid=kid, corr_id=corr_id)

        jwk = None
        # 1) Gather the key(s)
        try:
            jwk = await self._key_store.get_key(kid)
            keys = [jwk]
        except ValueError:
            if physical_path is None:
                logger.trace(
                    "handling_key_request_failed",
                    kid=kid,
                    corr_id=corr_id,
                    reason="key_not_found",
                )
                raise  # we really have nothing suitable
            keys = list(await self._key_store.get_keys_for_path(physical_path))

        assert jwk

        # 2) Build & sign the KeyAnnounce
        envelope_factory = self._envelope_factory
        if not envelope_factory:
            raise RuntimeError("Envelope factory not available - key manager not properly initialized")

        env = envelope_factory.create_envelope(
            frame=KeyAnnounceFrame(
                keys=keys,
                physical_path=jwk["physical_path"],
            ),
            corr_id=corr_id,
        )

        # 3) Send it the right way with proper delivery context
        node_id = self._node_id
        if not node_id:
            raise RuntimeError("Node ID not available - key manager not properly initialized")

        delivery_context = local_delivery_context(node_id)

        # Request stickiness for encryption keys
        if jwk.get("use") == "enc":
            delivery_context.stickiness_required = True
            delivery_context.sticky_sid = original_client_sid  # Set original client SID for AFT
            logger.debug(
                "key_announce_stickiness_set",
                kid=kid,
                corr_id=corr_id,
                key_use=jwk.get("use"),
                original_client_sid=original_client_sid,
            )

        if origin is DeliveryOriginType.DOWNSTREAM:
            if not self._routing_node:
                raise RuntimeError(
                    "Forward downstream not available - routing functionality not initialized"
                )
            # The child that asked lives on `from_seg`
            await self._routing_node.forward_to_route(from_seg, env, delivery_context)
        else:  # origin == UPSTREAM
            if not self._node:
                raise RuntimeError("Node not available - key manager not properly initialized")
            await self._node.forward_upstream(env, delivery_context)

    async def remove_keys_for_path(self, physical_path: str) -> int:
        """Remove all keys associated with the given physical path.

        This is useful when a system reconnects with new keys - we want to
        remove all old keys for that system before adding the new ones.

        Returns the number of keys removed.
        """
        removed_count = await self._key_store.remove_keys_for_path(physical_path)
        logger.debug(
            "removed_keys_for_path",
            physical_path=physical_path,
            removed_count=removed_count,
        )
        return removed_count

    async def get_keys_for_path(self, physical_path: str) -> Iterable[dict[str, Any]]:
        """Get all keys associated with the given physical path."""
        return await self._key_store.get_keys_for_path(physical_path)



"""Composite load balancing strategy implementation."""

from typing import Any, Optional, Sequence

from naylence.fame.core import FameEnvelope
from naylence.fame.util.logging import getLogger

from .load_balancing_strategy import LoadBalancingStrategy

logger = getLogger(__name__)


class CompositeLoadBalancingStrategy(LoadBalancingStrategy):
    """
    Composite load balancing strategy that chains multiple strategies with fallback.

    Tries each strategy in order until one returns a non-None result.
    This allows for sophisticated routing policies like "sticky first, then HRW fallback".
    """

    def __init__(self, strategies: Sequence[LoadBalancingStrategy]):
        if not strategies:
            raise ValueError("CompositeLoadBalancingStrategy requires at least one strategy")
        self.strategies = list(strategies)

    def choose(self, pool_key: Any, segments: Sequence[str], envelope: FameEnvelope) -> Optional[str]:
        """
        Try each strategy in order until one returns a non-None result.

        Args:
            pool_key: The pool identifier
            segments: Available segments in the pool
            envelope: The envelope being routed

        Returns:
            The chosen segment ID from the first successful strategy, or None if all fail
        """
        if not segments:
            return None

        for i, strategy in enumerate(self.strategies):
            try:
                result = strategy.choose(pool_key, segments, envelope)
                if result is not None:
                    logger.debug(
                        "composite_strategy_success",
                        envelope_id=envelope.id,
                        pool_key=pool_key,
                        strategy_index=i,
                        strategy_type=type(strategy).__name__,
                        result=result,
                    )
                    return result
            except Exception as e:
                logger.warning(
                    "composite_strategy_error",
                    envelope_id=envelope.id,
                    pool_key=pool_key,
                    strategy_index=i,
                    strategy_type=type(strategy).__name__,
                    error=str(e),
                )
                # Continue to next strategy on error

        logger.debug(
            "composite_strategy_all_failed",
            envelope_id=envelope.id,
            pool_key=pool_key,
            strategy_count=len(self.strategies),
        )
        return None



from __future__ import annotations

from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING, Any, Optional

from naylence.fame.core import (
    AuthorizationContext,
    FameDeliveryContext,
    FameEnvelope,
    NodeAttachFrame,
    generate_id,
)
from naylence.fame.node.node_context import FameNodeAuthorizationContext
from naylence.fame.node.node_event_listener import NodeEventListener
from naylence.fame.node.node_like import NodeLike
from naylence.fame.security.auth.authorizer import Authorizer
from naylence.fame.security.auth.token_issuer import TokenIssuer
from naylence.fame.security.auth.token_verifier import TokenVerifier
from naylence.fame.security.auth.token_verifier_provider import TokenVerifierProvider
from naylence.fame.util.logging import getLogger

if TYPE_CHECKING:
    pass

logger = getLogger(__name__)


class OAuth2Authorizer(NodeEventListener, Authorizer, TokenVerifierProvider):
    """
    OAuth2-specific authorizer for general-purpose OAuth2 providers.

    Unlike DefaultAuthorizer, this authorizer:
    - Only validates standard OAuth2 claims (audience, scopes, issuer, expiration)
    - Doesn't require custom claims like system_id, instance_id mapping
    - Suitable for Auth0, Okta, and other general-purpose OAuth2 providers
    - Uses the node attach frame values as the source of truth for system/instance IDs
    - Supports optional audience validation (defaults to node physical path if not specified)
    """

    def __init__(
        self,
        *,
        token_verifier: TokenVerifier,
        token_issuer: Optional[TokenIssuer] = None,
        audience: Optional[str] = None,
        required_scopes: Optional[list[str]] = None,
        require_scope: bool = True,
        default_ttl_sec: int = 3600,
        max_ttl_sec: int = 86400,
        reverse_auth_ttl_sec: int = 86400,
        enforce_token_subject_node_identity: bool = False,
        trusted_client_scope: str = "node.trusted",
    ):
        super().__init__()
        self._token_verifier = token_verifier
        self._audience = audience
        self._required_scopes = set(required_scopes) if required_scopes else set()
        self._require_scope = require_scope
        self._default_ttl_sec = default_ttl_sec
        self._max_ttl_sec = max_ttl_sec
        self._token_issuer = token_issuer
        self._reverse_auth_ttl_sec = reverse_auth_ttl_sec
        self._enforce_token_subject_node_identity = enforce_token_subject_node_identity
        self._trusted_client_scope = trusted_client_scope
        self._node: Optional[NodeLike] = None

    async def on_node_started(self, node: NodeLike) -> None:
        self._node = node

    @property
    def token_verifier(self) -> TokenVerifier:
        """Get the welcome token verifier used by this authorizer."""
        return self._token_verifier

    def create_reverse_authorization_config(self, node: NodeLike) -> Optional[Any]:
        """
        Create a bearer token configuration for reverse connections using TokenIssuer.

        Generates a token using the same issuer that would be used for welcome tokens,
        ensuring consistency in token format and validation.

        Returns:
            BearerTokenHeaderAuth instance or None if not configured
        """
        if not self._token_issuer:
            return None

        # Calculate expiration time
        expires_at = datetime.now(timezone.utc) + timedelta(seconds=self._reverse_auth_ttl_sec)

        try:
            # Generate token using the welcome token issuer
            # We use a special system_id pattern to indicate this is for reverse connections
            token = self._token_issuer.issue(
                claims={
                    "iss": self._token_issuer.issuer,
                    "aud": self._audience or node.physical_path,  # Use node physical path as audience
                    "exp": int(expires_at.timestamp()),
                    "sub": f"reverse-auth-{node.id}",  # Unique subject for reverse auth
                    "instance_id": getattr(node, "instance_id", None),  # Include instance ID
                    "capabilities": (list(self._required_scopes) if self._required_scopes else []),
                }
            )

            logger.debug(
                "reverse_authorization_token_generated",
                node_id=node.id,
                expires_at=expires_at.isoformat(),
                capabilities=(list(self._required_scopes) if self._required_scopes else ["fame:connect"]),
            )

            # Create the proper Auth instance
            from naylence.fame.security.auth.bearer_token_header_auth_injection_strategy_factory import (
                BearerTokenHeaderAuthInjectionStrategyConfig,
            )
            from naylence.fame.security.auth.static_token_provider_factory import (
                StaticTokenProviderConfig,
            )

            static_token_config = StaticTokenProviderConfig(
                type="StaticTokenProvider", token=token, expires_at=expires_at
            )

            return BearerTokenHeaderAuthInjectionStrategyConfig(token_provider=static_token_config)

        except Exception as e:
            logger.warning("failed_to_generate_reverse_auth_token", node_id=node.id, error=str(e))
            return None

    async def authenticate(
        self,
        credentials: str | bytes,
    ) -> Optional[AuthorizationContext]:
        """Authenticate using OAuth2 JWT token."""
        # Extract token from credentials
        token = credentials
        if isinstance(credentials, bytes):
            token = credentials.decode("utf-8")

        # Strip "Bearer " prefix if present
        if isinstance(token, str) and token.lower().startswith("bearer "):
            token = token[7:]

        if not token:
            return None

        try:
            # Use configured audience, or fall back to target node physical path
            expected_audience = self._audience or (self._node.physical_path if self._node else None)
            raw_claims = await self._token_verifier.verify(
                str(token),  # Ensure token is string
                expected_audience=expected_audience,
            )
        except Exception as e:
            logger.warning("token_verification_failed", exc_info=e)
            return None

        # Validate scopes if required
        token_scopes = self._extract_scopes_from_claims(raw_claims)
        if self._require_scope and self._required_scopes:
            if not self._required_scopes.intersection(token_scopes):
                logger.warning(
                    f"Token scopes {token_scopes} do not include any of required scopes {
                        self._required_scopes
                    }"
                )
                return None

        # Create authorization context
        return AuthorizationContext(
            authenticated=True,
            principal=raw_claims.get("sub"),
            claims=raw_claims,
            granted_scopes=list(token_scopes),
            auth_method="oauth2_jwt",
        )

    async def authorize(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Authorize access based on the current authentication context and operation.
        """

        if context and context.security and context.security.authorization:
            auth_context = context.security.authorization
        else:
            return None

        if auth_context.authorized is True:
            return auth_context

        # Check scopes if required and envelope is provided
        if self._require_scope and envelope:
            granted_scopes = set(auth_context.granted_scopes) if auth_context.granted_scopes else set()
            if self._required_scopes and not self._required_scopes.intersection(granted_scopes):
                return None

        # For OAuth2, if authenticated and scope checks pass, then authorized
        return auth_context

    def _extract_scopes_from_claims(self, claims: dict) -> set[str]:
        """Extract scopes from JWT claims. Handles both 'scope' and 'scopes' fields."""
        scopes = set()

        # Handle 'scope' field (space-separated string - OAuth2 standard)
        if "scope" in claims:
            scope_str = claims["scope"]
            if isinstance(scope_str, str):
                scopes.update(scope_str.split())

        # Handle 'scopes' field (array - some providers use this)
        if "scopes" in claims:
            scope_list = claims["scopes"]
            if isinstance(scope_list, list):
                scopes.update(scope_list)

        return scopes

    async def validate_node_attach_request(
        self,
        node: NodeLike,
        frame: NodeAttachFrame,
        auth_context: Optional[AuthorizationContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Validate node attach request using the already-authenticated context.
        No longer uses attach_token from frame - relies on wire-level auth.
        """
        if not auth_context or not auth_context.authenticated:
            return None  # Must be authenticated via wire-level auth first

        # Extract claims from the auth context (set during authenticate())
        if not hasattr(auth_context, "claims") or not auth_context.claims:
            return None

        claims = auth_context.claims

        # Validate required scopes if scope checking is enabled
        token_scopes = self._extract_scopes_from_claims(claims)
        if self._require_scope and self._required_scopes:
            if not any(scope in token_scopes for scope in self._required_scopes):
                logger.warning(
                    "oauth2_attach_missing_required_scope",
                    required_scopes=list(self._required_scopes),
                    token_scopes=list(token_scopes),
                )
                return None

        # Enforce token subject node identity if enabled and not a trusted client
        if self._enforce_token_subject_node_identity:
            is_trusted_client = self._trusted_client_scope in token_scopes
            if is_trusted_client:
                logger.debug(
                    "oauth2_attach_trusted_client_bypass",
                    system_id=frame.system_id,
                    trusted_scope=self._trusted_client_scope,
                )
            else:
                validation_result = self._validate_token_subject_node_identity(
                    frame.system_id, claims
                )
                if not validation_result:
                    return None

        # Create node-specific authorization context
        return FameNodeAuthorizationContext(
            sub=claims.get("sub", frame.system_id),
            aud=node.id,
            instance_id=frame.instance_id,
            assigned_path=frame.assigned_path,
            accepted_capabilities=(list(frame.capabilities) if frame.capabilities else None),
            accepted_logicals=(list(frame.accepted_logicals) if frame.accepted_logicals else None),
            # Include additional OAuth2 claims
            scopes=list(token_scopes),
            claims=claims,
        )

    def _validate_token_subject_node_identity(
        self,
        system_id: str,
        claims: dict,
    ) -> bool:
        """
        Validate that the node's system_id is prefixed with a hash of the token subject.

        This enforces that nodes using OAuth2 authentication have identities
        that are cryptographically bound to their token subject claim.
        """
        sub = claims.get("sub")

        if not isinstance(sub, str) or not sub.strip():
            logger.warning(
                "oauth2_attach_missing_subject_claim",
                system_id=system_id,
            )
            return False

        expected_prefix = generate_id(
            mode="fingerprint",
            material=sub,
            length=8,
        )

        if not system_id.startswith(f"{expected_prefix}-"):
            logger.warning(
                "oauth2_attach_node_identity_mismatch",
                system_id=system_id,
                expected_prefix=expected_prefix,
                subject=sub,
            )
            return False

        logger.debug(
            "oauth2_attach_node_identity_verified",
            system_id=system_id,
            expected_prefix=expected_prefix,
        )

        return True



"""
Connector Selection Policy for choosing the appropriate connector type when handling NodeAttach frames.

This module addresses the architectural concerns about hardcoded connector selection
by providing a pluggable policy system that considers:
- Client's supported inbound connectors
- Node preferences
- Inbound connector type that received the request
- Fallback strategies
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Protocol

from naylence.fame.grants.connection_grant import ConnectionGrant
from naylence.fame.grants.grant import GRANT_PURPOSE_NODE_ATTACH
from naylence.fame.grants.http_connection_grant import HttpConnectionGrant
from naylence.fame.util.logging import getLogger

if TYPE_CHECKING:
    from naylence.fame.core import NodeAttachFrame
    from naylence.fame.node.routing_node_like import RoutingNodeLike

logger = getLogger(__name__)


class ConnectorType(Enum):
    """Known connector types."""

    HTTP_STATELESS = "HttpStatelessConnector"
    WEBSOCKET_STATELESS = "WebSocketStatelessConnector"
    WEBSOCKET = "WebSocketConnector"


@dataclass
class GrantSelectionContext:
    """Context information for connector selection decisions."""

    child_id: str
    attach_frame: NodeAttachFrame
    callback_grant_type: str  # Type of connector that received the attach request
    node: RoutingNodeLike

    @property
    def client_supported_callback_grants(self) -> List[Dict[str, Any]]:
        """Get the list of connectors supported by the client."""
        if not self.attach_frame.callback_grants:
            return []

        # Convert any connector config objects to dicts for uniform processing
        result = []
        for connector in self.attach_frame.callback_grants:
            if isinstance(connector, dict):
                result.append(connector)
            else:
                # Assume it's a ConnectorConfig with model_dump method
                result.append(connector.model_dump(by_alias=True))
        return result


class GrantSelectionResult:
    """Result of connector selection containing the chosen config and metadata."""

    def __init__(
        self,
        grant: ConnectionGrant,
        selection_reason: str,
        fallback_used: bool = False,
    ):
        self.grant = grant
        self.selection_reason = selection_reason
        self.fallback_used = fallback_used

    def __repr__(self) -> str:
        return (
            f"GrantSelectionResult("
            f"type={self.grant.type}, "
            f"reason='{self.selection_reason}', "
            f"fallback={self.fallback_used})"
        )


class GrantSelectionStrategy(Protocol):
    """Protocol for connector selection strategies."""

    def select_callback_grant(self, context: GrantSelectionContext) -> Optional[GrantSelectionResult]:
        """
        Select a connector configuration based on the context.

        Returns:
            GrantSelectionResult if a suitable connector is found, None otherwise
        """
        ...


class GrantSelectionPolicy:
    """
    Main policy class that orchestrates connector selection using pluggable strategies.

    This addresses the TODO comments by providing a flexible, configurable approach
    to connector selection that considers client preferences, node policies, and
    fallback strategies.
    """

    def __init__(self, strategies: Optional[List[GrantSelectionStrategy]] = None):
        self.strategies = strategies or [
            PreferSameTypeStrategy(),
            PreferHttpStrategy(),
            ClientPreferenceStrategy(),
        ]

    def select_callback_grant(self, context: GrantSelectionContext) -> Optional[GrantSelectionResult]:
        """
        Select the best connector for the given context.

        Iterates through strategies until one returns a result.
        Raises ValueError if no suitable connector can be found.
        """
        logger.debug(
            "selecting_connector",
            child=context.child_id,
            inbound_type=context.callback_grant_type,
            client_grants=[c.get("type") for c in context.client_supported_callback_grants],
        )

        for strategy in self.strategies:
            result = strategy.select_callback_grant(context)
            if result:
                logger.debug(
                    "connector_selected",
                    child=context.child_id,
                    selected_type=result.grant.type,
                    strategy=strategy.__class__.__name__,
                    reason=result.selection_reason,
                    fallback=result.fallback_used,
                )
                return result

        # No suitable connector found - raise error with detailed information
        supported_types = [c.get("type") for c in context.client_supported_callback_grants]
        error_msg = (
            f"No suitable connector found for child {context.child_id}. "
            f"Client supports: {supported_types}, "
            f"inbound type: {context.callback_grant_type}"
        )

        logger.warning(
            "connector_selection_failed",
            child=context.child_id,
            client_connectors=supported_types,
            inbound_type=context.callback_grant_type,
            reason="No matching strategy found",
        )

        raise ValueError(error_msg)


class PreferSameTypeStrategy(GrantSelectionStrategy):
    """Strategy that prefers to use the same connector type as the inbound connection."""

    def select_callback_grant(self, context: GrantSelectionContext) -> Optional[GrantSelectionResult]:
        """Select connector matching the inbound connector type if available."""
        target_type = context.callback_grant_type

        for grant_dict in context.client_supported_callback_grants:
            if grant_dict.get("type") == target_type:
                config = self._crate_grant_from_dict(grant_dict)
                if config:
                    return GrantSelectionResult(
                        grant=config,
                        selection_reason=f"Matching inbound connector type: {target_type}",
                    )

        return None

    def _crate_grant_from_dict(self, grant_dict: Dict[str, Any]) -> Optional[ConnectionGrant]:
        """Create a connector config from a dictionary representation."""
        grant_type = grant_dict.get("type")

        if grant_type == "HttpConnectionGrant":
            return self._create_http_config(grant_dict)
        elif grant_type == "WebSocketConnectionGrant":
            return self._create_websocket_config(grant_dict)

        return None

    def _create_http_config(self, connector_dict: Dict[str, Any]) -> Optional[ConnectionGrant]:
        """Create HTTP connector config from dict."""

        # Extract URL from the dict
        url = connector_dict.get("url") or connector_dict.get("params", {}).get("url")
        if not url:
            return None

        return HttpConnectionGrant(
            url=url,
            purpose=connector_dict.get("purpose", GRANT_PURPOSE_NODE_ATTACH),
            # max_queue=connector_dict.get("max_queue", 1024),
            auth=connector_dict.get("auth"),
        )

    def _create_websocket_config(self, connector_dict: Dict[str, Any]) -> Optional[ConnectionGrant]:
        """Create WebSocket connector config from dict."""

        params = connector_dict.get("params", {})
        if not params:
            return None

        from naylence.fame.grants.websocket_connection_grant import WebSocketConnectionGrant

        return WebSocketConnectionGrant(
            type=connector_dict.get("type", "WebSocketConnectionGrant"),
            purpose=connector_dict.get("purpose", GRANT_PURPOSE_NODE_ATTACH),
            # params=params,
            auth=connector_dict.get("auth"),
        )


class PreferHttpStrategy(GrantSelectionStrategy):
    """Strategy that prefers HTTP connectors when available."""

    def select_callback_grant(self, context: GrantSelectionContext) -> Optional[GrantSelectionResult]:
        """Select HTTP connector if available."""
        for connector_dict in context.client_supported_callback_grants:
            if connector_dict.get("type") == "HttpConnectionGrant":
                config = self._create_http_config(connector_dict)
                if config:
                    return GrantSelectionResult(
                        grant=config,
                        selection_reason="Preferred HTTP connector type",
                    )

        return None

    def _create_http_config(self, connector_dict: Dict[str, Any]) -> Optional[ConnectionGrant]:
        """Create HTTP connector config from dict."""

        url = connector_dict.get("url") or connector_dict.get("params", {}).get("url")
        if not url:
            return None

        return HttpConnectionGrant(
            url=url,
            purpose=connector_dict.get("purpose", GRANT_PURPOSE_NODE_ATTACH),
            # max_queue=connector_dict.get("max_queue", 1024),
            auth=connector_dict.get("auth"),
        )


class ClientPreferenceStrategy(GrantSelectionStrategy):
    """Strategy that uses the first connector provided by the client."""

    def select_callback_grant(self, context: GrantSelectionContext) -> Optional[GrantSelectionResult]:
        """Select the first available connector from the client's list."""
        if not context.client_supported_callback_grants:
            return None

        first_connector = context.client_supported_callback_grants[0]
        config = self._create_config_from_dict(first_connector)

        if config:
            return GrantSelectionResult(
                grant=config,
                selection_reason=f"Client's first preference: {first_connector.get('type')}",
            )

        return None

    def _create_config_from_dict(self, grant_dict: Dict[str, Any]) -> Optional[ConnectionGrant]:
        """Create a connector config from a dictionary representation."""
        # Reuse the logic from PreferSameTypeStrategy
        strategy = PreferSameTypeStrategy()
        return strategy._crate_grant_from_dict(grant_dict)


# Default policy instance
default_grant_selection_policy = GrantSelectionPolicy()



"""
Factory for TraceEmitter implementations following Fame's ResourceFactory pattern.
"""

from __future__ import annotations

from typing import Any, Optional, TypeVar

from naylence.fame.factory import ResourceConfig, ResourceFactory, create_resource
from naylence.fame.telemetry.trace_emitter import TraceEmitter


class TraceEmitterConfig(ResourceConfig):
    """Base configuration for TraceEmitter implementations."""

    type: str = "TraceEmitter"


C = TypeVar("C", bound=TraceEmitterConfig)


class TraceEmitterFactory(ResourceFactory[TraceEmitter, C]):
    """Abstract factory for creating TraceEmitter instances."""

    async def create(self, config: Optional[C | dict[str, Any]] = None, **kwargs: Any) -> TraceEmitter:
        """Create a TraceEmitter instance."""
        raise NotImplementedError

    @classmethod
    async def create_trace_emitter(
        cls,
        cfg: Optional[C | dict[str, Any]] = None,
        **kwargs,
    ) -> Optional[TraceEmitter]:
        """Create a TraceEmitter instance based on the provided configuration."""

        return await create_resource(
            TraceEmitterFactory,
            cfg,
            **kwargs,
        )



from __future__ import annotations

from naylence.fame.security.util import require_crypto
from naylence.fame.util.crypto_util import DevKeyPair, generate_key_pair_and_jwks


def create_ed25519_keypair(kid: str = "dev") -> DevKeyPair:
    require_crypto()
    from cryptography.hazmat.primitives.asymmetric import ed25519

    data = generate_key_pair_and_jwks(
        kid=kid,
        key_gen_fn=ed25519.Ed25519PrivateKey.generate,
        algorithm="EdDSA",
    )
    data["jwks"]["keys"][0]["kid"] = kid
    return DevKeyPair(**data)



from __future__ import annotations

from typing import Protocol, runtime_checkable

from naylence.fame.security.auth.token_verifier import TokenVerifier


@runtime_checkable
class TokenVerifierProvider(Protocol):
    """
    Protocol for authorizers that can provide their internal token verifier.

    This allows reusing the same verifier instance that the authorizer uses
    for token validation, avoiding duplicate configuration and setup.
    """

    @property
    def token_verifier(self) -> TokenVerifier:
        """Get the token verifier used by this authorizer."""
        ...



"""
Factory system for creating authentication injection strategies.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from pydantic import ConfigDict
from pydantic.alias_generators import to_camel

from naylence.fame.factory import ResourceConfig, ResourceFactory, create_resource

if TYPE_CHECKING:
    from naylence.fame.security.auth.auth_injection_strategy import AuthInjectionStrategy


class AuthInjectionStrategyConfig(ResourceConfig):
    """Base class for connector authentication configurations."""

    model_config = ConfigDict(
        alias_generator=to_camel,
        populate_by_name=True,
        extra="allow",
        arbitrary_types_allowed=True,
    )


C = TypeVar("C", bound=AuthInjectionStrategyConfig)


class AuthInjectionStrategyFactory(ResourceFactory):
    """Base factory for creating auth injection strategies."""

    @staticmethod
    async def create_auth_strategy(auth_config: AuthInjectionStrategyConfig) -> AuthInjectionStrategy:
        return await create_resource(AuthInjectionStrategyFactory, auth_config)



"""
NoSecurityManager - A security manager that provides no security components.

This implementation provides a completely open security posture with no encryption,
signing, verification, or other security features. It's intended for development,
testing, or scenarios where security is handled at a different layer.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from naylence.fame.security.auth.authorizer import Authorizer
from naylence.fame.security.policy import NoSecurityPolicy
from naylence.fame.security.security_manager import SecurityManager
from naylence.fame.util.logging import getLogger

if TYPE_CHECKING:
    from naylence.fame.node.envelope_security_handler import EnvelopeSecurityHandler
    from naylence.fame.node.node_like import NodeLike
    from naylence.fame.node.secure_channel_frame_handler import (
        SecureChannelFrameHandler,
    )
    from naylence.fame.security.cert.certificate_manager import CertificateManager
    from naylence.fame.security.encryption.encryption_manager import EncryptionManager
    from naylence.fame.security.keys.key_manager import KeyManager
    from naylence.fame.security.policy import SecurityPolicy
    from naylence.fame.security.signing.envelope_signer import EnvelopeSigner
    from naylence.fame.security.signing.envelope_verifier import EnvelopeVerifier

logger = getLogger(__name__)


class NoSecurityManager(SecurityManager):
    """
    A security manager implementation that provides no security features.

    This implementation returns None for all security components and provides
    no-op implementations for all event handlers. It's suitable for scenarios
    where no security is required or security is handled at a different layer.
    """

    def __init__(self):
        """Initialize the no-security manager with a NoSecurityPolicy and minimal components."""
        self._policy = NoSecurityPolicy()

        # Create a minimal no-op authorizer to satisfy Sentinel requirements
        # This authorizer allows all connections without any actual security checks
        try:
            from naylence.fame.security.auth.noop_authorizer import NoopAuthorizer

            self._authorizer = NoopAuthorizer()
        except ImportError:
            # Fallback if NoopAuthorizer doesn't exist
            self._authorizer = None

    @property
    def policy(self) -> SecurityPolicy:
        """Get the no-security policy."""
        return self._policy

    @property
    def envelope_signer(self) -> Optional[EnvelopeSigner]:
        """Get the envelope signer (always None)."""
        return None

    @property
    def envelope_verifier(self) -> Optional[EnvelopeVerifier]:
        """Get the envelope verifier (always None)."""
        return None

    @property
    def encryption(self) -> Optional[EncryptionManager]:
        """Get the encryption manager (always None)."""
        return None

    @property
    def key_manager(self) -> Optional[KeyManager]:
        """Get the key manager (always None)."""
        return None

    @property
    def supports_overlay_security(self) -> bool:
        return False

    @property
    def authorizer(self) -> Optional[Authorizer]:
        """Get the node attach authorizer (minimal no-op authorizer)."""
        return self._authorizer

    @property
    def certificate_manager(self) -> Optional[CertificateManager]:
        """Get the certificate manager (always None)."""
        return None

    @property
    def envelope_security_handler(self) -> Optional[EnvelopeSecurityHandler]:
        """Get the envelope security handler (always None)."""
        return None

    @property
    def secure_channel_frame_handler(self) -> Optional[SecureChannelFrameHandler]:
        """Get the channel frame handler (always None)."""
        return None

    def get_encryption_key_id(self) -> Optional[str]:
        """Get the encryption key ID (always None)."""
        return None

    def get_shareable_keys(self) -> Any:
        """Get keys to provide to child nodes (always None for no security)."""
        return None

    # NodeEventListener implementation - all no-op

    async def on_node_initialized(self, node: NodeLike) -> None:
        """Handle node initialization (no-op)."""
        logger.debug("noop_security_manager_node_initialized", node_id=getattr(node, "id", None))

    async def on_node_started(self, node: NodeLike) -> None:
        """Handle node start (no-op)."""
        logger.debug("noop_security_manager_node_started", node_id=getattr(node, "id", None))

    async def on_node_stopped(self, node: NodeLike) -> None:
        """Handle node stop (no-op)."""
        logger.debug("noop_security_manager_node_stopped", node_id=getattr(node, "id", None))

    async def on_welcome(self, welcome_frame: Any) -> None:
        """Handle child welcome event (no-op)."""
        logger.debug("noop_security_manager_child_welcome")

    async def on_heartbeat_received(self, envelope: Any) -> None:
        """Handle heartbeat received event (no-op)."""
        logger.debug("noop_security_manager_heartbeat_received")

    async def on_child_attach(
        self,
        *,
        child_system_id: str,
        child_keys: Any,
        node_like: NodeLike,
        origin_type: Any = None,
        assigned_path: Optional[str] = None,
        old_assigned_path: Optional[str] = None,
        is_rebind: bool = False,
    ) -> None:
        """Handle child attachment (no-op for no security)."""
        logger.debug("noop_security_manager_child_attach", child_system_id=child_system_id)



"""
Authorization policy interface.

Defines the protocol for authorization policies that evaluate
whether a request should be allowed or denied.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, Literal, Optional, Protocol, runtime_checkable

if TYPE_CHECKING:
    from naylence.fame.core import FameDeliveryContext, FameEnvelope
    from naylence.fame.node.node_like import NodeLike

from .authorization_policy_definition import RuleAction

# The effect of an authorization decision
AuthorizationEffect = Literal["allow", "deny"]


@dataclass
class AuthorizationEvaluationStep:
    """
    Represents a single step in the policy evaluation process.

    Useful for debugging and auditing authorization decisions.
    """

    # Rule identifier that was evaluated
    rule_id: str

    # Result of the evaluation
    result: bool

    # Expression or condition that was evaluated
    expression: Optional[str] = None

    # Context values used in evaluation (for debugging)
    bound_values: Optional[dict[str, Any]] = None


@dataclass
class AuthorizationDecision:
    """The result of an authorization policy evaluation."""

    # The authorization effect: allow or deny
    effect: AuthorizationEffect

    # Human-readable reason for the decision
    reason: Optional[str] = None

    # Identifier of the rule that matched (for debugging/audit)
    matched_rule: Optional[str] = None

    # Evaluation trace for detailed debugging
    evaluation_trace: list[AuthorizationEvaluationStep] = field(default_factory=list)


@runtime_checkable
class AuthorizationPolicy(Protocol):
    """
    Interface for authorization policies that evaluate whether a request
    should be allowed or denied.

    The policy receives the same parameters as `Authorizer.authorize`,
    giving it full access to the node, envelope, and delivery context
    for making authorization decisions.
    """

    async def evaluate_request(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        context: Optional[FameDeliveryContext] = None,
        action: Optional[RuleAction] = None,
    ) -> AuthorizationDecision:
        """
        Evaluates an authorization request and returns a decision.

        Args:
            node: The node handling the request
            envelope: The FAME envelope being authorized
            context: Optional delivery context with authorization info, origin, etc.
            action: Optional authorization action token (route-oriented: Connect,
                    ForwardUpstream, ForwardDownstream, ForwardPeer, DeliverLocal, '*')

        Returns:
            A decision indicating whether to allow or deny the request
        """
        ...



from __future__ import annotations

from typing import TYPE_CHECKING, Optional

from naylence.fame.core import AuthorizationContext, FameDeliveryContext, FameEnvelope
from naylence.fame.core.protocol.frames import NodeAttachFrame
from naylence.fame.node.node_like import NodeLike
from naylence.fame.security.auth.authorizer import Authorizer, RouteAuthorizationResult

if TYPE_CHECKING:
    from naylence.fame.security.auth.policy.authorization_policy_definition import RuleAction


class NoopAuthorizer(Authorizer):
    """
    A no-op NodeAttachAuthorizer that allows all attach requests unconditionally.
    """

    async def authenticate(
        self,
        credentials: str | bytes,
    ) -> Optional[AuthorizationContext]:
        # No authentication needed, just return an empty context
        return AuthorizationContext()

    async def authorize(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Authorize access for any envelope unconditionally.

        Supports both new and legacy interfaces:
        - New: authorize(node, envelope, context: FameDeliveryContext)
        - Legacy: authorize(node, envelope, auth_context: AuthorizationContext)

        For NoopAuthorizer, always allows access by returning a valid authorization context.
        """
        # Backward compatibility: detect if context is actually an AuthorizationContext
        if context is None:
            # Return a basic authorization context
            return AuthorizationContext()
        elif hasattr(context, "authenticated"):
            # Legacy interface: context is an AuthorizationContext (duck typing check)
            return context  # type: ignore
        else:
            # New interface: context is a FameDeliveryContext
            # Extract existing authorization context from delivery context if available
            if context and context.security and context.security.authorization:
                return context.security.authorization

            # Return a basic authorization context
            return AuthorizationContext()

    async def authorize_route(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        action: RuleAction,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[RouteAuthorizationResult]:
        """
        Authorize routing action unconditionally.

        For NoopAuthorizer, always allows the routing action.
        """
        return RouteAuthorizationResult(authorized=True)

    # Legacy method for backward compatibility with NodeAttach frames
    async def validate_node_attach_request(
        self,
        node: NodeLike,
        frame: NodeAttachFrame,
        auth_context: Optional[AuthorizationContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Legacy method for NodeAttach frame validation.
        Always returns a valid context since this is a no-op authorizer.
        """
        return AuthorizationContext()



from __future__ import annotations

from typing import Optional

from .credential_provider import CredentialProvider


class NoneCredentialProvider(CredentialProvider):
    """
    A credential provider that returns None for all requests.
    Good for "no-auth" scenarios.
    """

    async def get(self) -> Optional[str]:
        """Always returns None (no credentials)."""
        return None



"""Sticky load balancing strategy implementation."""

from typing import Any, Optional, Sequence

from naylence.fame.core import FameEnvelope
from naylence.fame.stickiness.load_balancer_stickiness_manager import (
    LoadBalancerStickinessManager,
)
from naylence.fame.util.logging import getLogger

from .load_balancing_strategy import LoadBalancingStrategy

logger = getLogger(__name__)


class StickyLoadBalancingStrategy(LoadBalancingStrategy):
    """
    Load balancing strategy that provides session affinity by delegating to a SentinelStickinessManager.

    This strategy delegates all stickiness logic to the configured SentinelStickinessManager:
    1. For inbound envelopes, asks the manager to resolve any sticky associations
       (AFT, SID cache, deterministic SID)
    2. If the manager returns a replica ID, routes to that replica
    3. If no sticky match, returns None to allow composite strategy fallback
    """

    def __init__(self, stickiness_manager: LoadBalancerStickinessManager):
        self.stickiness_manager = stickiness_manager
        self._last_chosen_replica: Optional[str] = None

    def choose(self, pool_key: Any, segments: Sequence[str], envelope: FameEnvelope) -> Optional[str]:
        """
        Choose a segment from the pool based on session affinity.

        Args:
            pool_key: The pool identifier
            segments: Available segments in the pool
            envelope: The envelope being routed

        Returns:
            The sticky segment ID if AFT match found, or None to allow fallback
        """
        if not segments:
            return None

        # Ask the stickiness manager to resolve any sticky association (AFT/SID cache/deterministic SID)
        sticky_replica_segment = self.stickiness_manager.get_sticky_replica_segment(envelope, segments)

        if sticky_replica_segment and sticky_replica_segment in segments:
            logger.debug(
                "routing_via_stickiness",
                envelope_id=envelope.id,
                pool_key=pool_key,
                replica_id=sticky_replica_segment,
                aft_present=bool(envelope.aft),
                sid_present=bool(envelope.sid),
            )
            self._last_chosen_replica = sticky_replica_segment
            return sticky_replica_segment

        # No stickiness match - return None to allow composite strategy fallback
        logger.debug(
            "no_stickiness_match_fallback",
            envelope_id=envelope.id,
            pool_key=pool_key,
            aft_present=bool(envelope.aft),
            sid_present=bool(envelope.sid),
        )
        return None

    def get_metrics(self):
        """Get stickiness metrics by delegating to the stickiness manager."""
        return getattr(self.stickiness_manager, "get_metrics", lambda: {})()

    def get_associations(self):
        """Get current AFT associations by delegating to the stickiness manager."""
        return getattr(self.stickiness_manager, "get_associations", lambda: {})()



from __future__ import annotations

from typing import Any, Optional

from naylence.fame.core import (
    create_resource,
)
from naylence.fame.sentinel.load_balancing.load_balancing_strategy_factory import (
    LoadBalancingStrategyConfig,
    LoadBalancingStrategyFactory,
)
from naylence.fame.sentinel.routing_policy import (
    RoutingPolicy,
    RoutingPolicyConfig,
    RoutingPolicyFactory,
)
from naylence.fame.util.logging import getLogger

logger = getLogger(__name__)


class HybridPathRoutingPolicyConfig(RoutingPolicyConfig):
    type: str = "HybridPathRoutingPolicy"
    load_balancing_strategy: Optional[LoadBalancingStrategyConfig] = None


class HybridPathRoutingPolicyFactory(RoutingPolicyFactory):
    async def create(
        self,
        config: Optional[HybridPathRoutingPolicyConfig | dict[str, Any]] = None,
        **kwargs: Any,
    ) -> RoutingPolicy:
        assert config
        from naylence.fame.sentinel.hybrid_path_routing_policy import (
            HybridPathRoutingPolicy,
        )

        if isinstance(config, dict):
            lb_config = config.get("load_balancing_strategy")
        else:
            lb_config = config.load_balancing_strategy
        load_balancing_strategy = await create_resource(LoadBalancingStrategyFactory, lb_config)
        return HybridPathRoutingPolicy(load_balancing_strategy)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any, Optional

import websockets

from naylence.fame.connector.base_async_connector import BaseAsyncConnector
from naylence.fame.core import AuthorizationContext
from naylence.fame.errors.errors import FameTransportClose
from naylence.fame.util import logging

# ---------------------------------------------------------------------------
# Optional FastAPI import – safe when FastAPI isn’t installed
# ---------------------------------------------------------------------------
try:  # pragma: no cover
    from fastapi import WebSocket as _FastAPIWebSocket
    from fastapi import WebSocketDisconnect as _FastAPIWebSocketDisconnect
    from fastapi.websockets import WebSocketState as _FastAPIWebSocketState
except ModuleNotFoundError:  # FastAPI not installed
    _FastAPIWebSocket = None  # type: ignore
    _FastAPIWebSocketDisconnect = ()  # type: ignore
    _FastAPIWebSocketState = None  # type: ignore

# For type-checkers (they run in a virtual env where FastAPI *is* present)
if TYPE_CHECKING:  # pragma: no cover
    from fastapi import WebSocket as _FastAPIWebSocket
    from fastapi import WebSocketDisconnect as _FastAPIWebSocketDisconnect
    from fastapi.websockets import WebSocketState as _FastAPIWebSocketState


logger = logging.getLogger(__name__)


class WebSocketConnector(BaseAsyncConnector):
    """
    A thin transport adapter that can work with either:

    • `websockets.client.WebSocketClientProtocol` (the std-lib style)
    • `fastapi.WebSocket` (server side)

    FastAPI is treated as an **optional** dependency.  If FastAPI is not
    installed, only the std-lib websocket path is available.
    """

    def __init__(
        self,
        websocket: Any,
        *,
        authorization_context: Optional[AuthorizationContext] = None,
        drain_timeout: float = 0.1,
    ):
        super().__init__(drain_timeout=drain_timeout)

        self.websocket = websocket
        self._is_fastapi = _FastAPIWebSocket is not None and isinstance(websocket, _FastAPIWebSocket)

        self._authorization_context: Optional[AuthorizationContext] = authorization_context

    # --------------------------------------------------------------------- #
    # low-level send/recv/close                                              #
    # --------------------------------------------------------------------- #
    async def _transport_send_bytes(self, data: bytes) -> None:
        try:
            if self._is_fastapi:
                await self.websocket.send_bytes(data)
            else:
                await self.websocket.send(data)
        except (_FastAPIWebSocketDisconnect, websockets.ConnectionClosed) as ws:  # type: ignore
            raise FameTransportClose(ws.code, getattr(ws, "reason", "") or "peer closed")

    async def _transport_receive(self) -> bytes:
        """
        Receive bytes from the websocket with enhanced error handling and cancellation safety.
        """
        try:
            # Validate websocket object before attempting to receive
            if self.websocket is None:
                raise FameTransportClose(1006, "WebSocket object is None")

            # Use a timeout to prevent hanging during shutdown scenarios
            # Increased to 3600s to match Uvicorn keep-alive and prevent idle disconnects
            receive_timeout = 30.0

            if self._is_fastapi:
                # Check if receive_bytes method exists and is callable
                receive_method = getattr(self.websocket, "receive_bytes", None)

                if not receive_method or not callable(receive_method):
                    raise FameTransportClose(1006, "FastAPI WebSocket receive_bytes method not available")

                result = self.websocket.receive_bytes()
                # logger.debug("fastapi_receive_result_type", result_type=type(result).__name__)

                # Additional validation to ensure we have an awaitable
                if not hasattr(result, "__await__") and not asyncio.iscoroutine(result):
                    logger.error(
                        "fastapi_receive_not_awaitable",
                        result_type=type(result).__name__,
                        result_str=str(result)[:100],
                    )
                    raise FameTransportClose(
                        1006,
                        f"FastAPI receive_bytes returned non-awaitable: {type(result)}",
                    )

                # Use asyncio.wait_for to add timeout protection and better cancellation handling
                try:
                    return await asyncio.wait_for(result, timeout=receive_timeout)
                except asyncio.TimeoutError:
                    # Check if we're being cancelled during timeout
                    current_task = asyncio.current_task()
                    if current_task and current_task.cancelled():
                        raise asyncio.CancelledError("Receive timeout during task cancellation")
                    raise FameTransportClose(1006, "FastAPI receive_bytes timed out")
                except asyncio.CancelledError:
                    logger.debug("fastapi_receive_cancelled")
                    raise
                except Exception as e:
                    # Check if this is the known WebSocket shutdown race condition
                    # (normal during task cancellation)
                    if "await wasn't used with future" in str(e):
                        logger.debug(
                            "websocket_shutdown_race_condition_handled",
                            note="Normal WebSocket close timing - converting to cancellation",
                            websocket_state=getattr(self.websocket, "client_state", "unknown"),
                        )
                        # Convert to cancellation error if this happened during shutdown
                        current_task = asyncio.current_task()
                        if current_task and current_task.cancelled():
                            raise asyncio.CancelledError(
                                "Converted await future error during cancellation"
                            ) from e
                    raise
            else:
                # Check if recv method exists and is callable
                recv_method = getattr(self.websocket, "recv", None)

                if not recv_method or not callable(recv_method):
                    raise FameTransportClose(1006, "WebSocket recv method not available")

                result = self.websocket.recv()

                # Additional validation to ensure we have an awaitable
                if not hasattr(result, "__await__") and not asyncio.iscoroutine(result):
                    logger.error(
                        "websockets_recv_not_awaitable",
                        result_type=type(result).__name__,
                        result_str=str(result)[:100],
                    )
                    raise FameTransportClose(1006, f"WebSocket recv returned non-awaitable: {type(result)}")

                # Use asyncio.wait_for for timeout protection and better cancellation handling
                try:
                    return await asyncio.wait_for(result, timeout=receive_timeout)
                except asyncio.TimeoutError:
                    # Check if we're being cancelled during timeout
                    current_task = asyncio.current_task()
                    if current_task and current_task.cancelled():
                        raise asyncio.CancelledError("Receive timeout during task cancellation")
                    raise FameTransportClose(1006, "WebSocket recv timed out")
                except asyncio.CancelledError:
                    logger.debug("websockets_recv_cancelled")
                    raise
                except Exception as e:
                    # Check if this is the problematic "await wasn't used with future" during cancellation
                    if "await wasn't used with future" in str(e):
                        logger.warning(
                            "websockets_recv_await_future_during_cancellation",
                            error=str(e),
                            websocket_state=getattr(self.websocket, "state", "unknown"),
                        )
                        # Convert to cancellation error if this happened during shutdown
                        current_task = asyncio.current_task()
                        if current_task and current_task.cancelled():
                            raise asyncio.CancelledError(
                                "Converted await future error during cancellation"
                            ) from e
                    raise
        except TypeError as e:
            if "await wasn't used with future" in str(e):
                logger.debug(
                    "websocket_shutdown_race_condition_detected",
                    websocket_type=type(self.websocket).__name__,
                    is_fastapi=self._is_fastapi,
                    note="Normal WebSocket close timing during shutdown",
                )

                # Check if current task is being cancelled - this is the expected case
                current_task = asyncio.current_task()
                if current_task and current_task.cancelled():
                    logger.debug(
                        "websocket_cancelled_during_receive",
                        task_name=current_task.get_name(),
                        note="Normal shutdown sequence",
                    )
                    raise asyncio.CancelledError("WebSocket cancelled during receive operation") from e

                # Try to get more info about what we're trying to await
                if self._is_fastapi:
                    try:
                        problematic_result = self.websocket.receive_bytes()
                        logger.error(
                            "fastapi_problematic_result",
                            result_type=type(problematic_result).__name__,
                            result_value=str(problematic_result)[:200],
                        )
                    except Exception as debug_e:
                        logger.error("fastapi_debug_error", error=str(debug_e))
                else:
                    try:
                        problematic_result = self.websocket.recv()
                        logger.error(
                            "websockets_problematic_result",
                            result_type=type(problematic_result).__name__,
                            result_value=str(problematic_result)[:200],
                        )
                    except Exception as debug_e:
                        logger.error("websockets_debug_error", error=str(debug_e))
            raise
        except (_FastAPIWebSocketDisconnect, websockets.ConnectionClosed) as ws:  # type: ignore
            raise FameTransportClose(ws.code, getattr(ws, "reason", "") or "peer closed")

    async def _transport_close(self, code: int, reason: str) -> None:
        if self._is_fastapi:
            # FastAPI's WebSocket has explicit state tracking
            if self.websocket.client_state == _FastAPIWebSocketState.CONNECTED:  # type: ignore
                try:
                    await self.websocket.close(code=code, reason=reason)  # type: ignore
                except Exception as e:  # type: ignore
                    logger.error("websocket_close_failed", error=e, exc_info=True)
                    # raise
        else:
            await self.websocket.close(code=code, reason=reason)

    @property
    def authorization_context(self) -> Optional[AuthorizationContext]:
        """Return the current authorization context."""
        return self._authorization_context

    @authorization_context.setter
    def authorization_context(self, context: Optional[AuthorizationContext]) -> None:
        """Set the authorization context for this connector."""
        self._authorization_context = context



from __future__ import annotations

from typing import TYPE_CHECKING, Any, List, Optional, TypeVar

from naylence.fame.factory import ResourceFactory, create_default_resource
from naylence.fame.security.encryption.encryption_manager import EncryptionManager
from naylence.fame.security.encryption.secure_channel_manager import (
    SecureChannelManager,
)
from naylence.fame.security.keys.attachment_key_validator import AttachmentKeyValidator
from naylence.fame.security.policy.security_policy import SecurityPolicy
from naylence.fame.security.security_manager_config import SecurityManagerConfig
from naylence.fame.security.signing.envelope_signer import EnvelopeSigner
from naylence.fame.security.signing.envelope_verifier import EnvelopeVerifier

from .security_manager import SecurityManager

if TYPE_CHECKING:
    from naylence.fame.node.node_event_listener import NodeEventListener
    from naylence.fame.security.auth.authorizer import Authorizer
    from naylence.fame.security.cert.certificate_manager import CertificateManager
    from naylence.fame.security.keys.key_manager import KeyManager

# Type variables for the factory pattern
C = TypeVar("C", bound=SecurityManagerConfig)


class SecurityManagerFactory(ResourceFactory[SecurityManager, C]):
    """Abstract ResourceFactory for creating SecurityManager instances."""

    async def create(self, config: Optional[C | dict[str, Any]] = None, **kwargs: Any) -> SecurityManager:
        """Create a SecurityManager instance."""
        raise NotImplementedError

    @classmethod
    async def create_security_manager(
        cls,
        policy: Optional[SecurityPolicy] = None,
        envelope_signer: Optional[EnvelopeSigner] = None,
        envelope_verifier: Optional[EnvelopeVerifier] = None,
        encryption_manager: Optional[EncryptionManager] = None,
        key_manager: Optional[KeyManager] = None,
        key_validator: Optional[AttachmentKeyValidator] = None,
        authorizer: Optional[Authorizer] = None,
        certificate_manager: Optional[CertificateManager] = None,
        secure_channel_manager: Optional[SecureChannelManager] = None,
        event_listeners: Optional[List[NodeEventListener]] = None,
    ) -> SecurityManager:
        from naylence.fame.security.security_manager_factory import (
            SecurityManagerFactory,
        )

        result = await create_default_resource(
            SecurityManagerFactory,
            config=None,
            policy=policy,
            envelope_signer=envelope_signer,
            envelope_verifier=envelope_verifier,
            encryption_manager=encryption_manager,
            key_manager=key_manager,
            key_validator=key_validator,
            authorizer=authorizer,
            certificate_manager=certificate_manager,
            secure_channel_manager=secure_channel_manager,
            event_listeners=event_listeners,
        )

        assert result is not None, "Failed to create SecurityManager instance"

        return result



from __future__ import annotations

from abc import ABC, abstractmethod


class SessionManager(ABC):
    """
    Abstract base class for session managers that handle node lifecycle operations.

    Session managers are responsible for managing the lifecycle of node connections,
    whether to upstream parents (UpstreamSessionManager) or for root sentinels
    (RootSessionManager).
    """

    @abstractmethod
    async def start(self, *, wait_until_ready: bool = True) -> None:
        """
        Start the session manager and begin managing the session lifecycle.

        Args:
            wait_until_ready: Whether to wait until the session is fully established
                            before returning. If True, this method will block until
                            the session is ready for use.
        """
        pass

    @abstractmethod
    async def stop(self) -> None:
        """
        Stop the session manager and clean up all resources.

        This should gracefully shut down any active connections, cancel background
        tasks, and release all resources.
        """
        pass



from __future__ import annotations

from typing import Optional

from naylence.fame.core import (
    AuthorizationContext,
    FameDeliveryContext,
    FameEnvelope,
    NodeAttachFrame,
)
from naylence.fame.node.node_context import FameNodeAuthorizationContext
from naylence.fame.node.node_like import NodeLike
from naylence.fame.security.auth.authorizer import Authorizer
from naylence.fame.security.credential import CredentialProvider


class SharedSecretAuthorizer(Authorizer):
    """
    A shared secret authorizer that validates tokens using a shared secret.
    """

    def __init__(self, credential_provider: CredentialProvider):
        self._credential_provider = credential_provider

    async def authenticate(
        self,
        credentials: str | bytes,
    ) -> Optional[AuthorizationContext]:
        """Validate shared secret credentials."""
        expected_secret = await self._credential_provider.get()
        if not expected_secret:
            raise ValueError("Shared secret not configured")

        # Extract token from credentials
        token = credentials
        if isinstance(credentials, bytes):
            token = credentials.decode("utf-8")

        # Strip "Bearer " prefix if present
        if isinstance(token, str) and token.lower().startswith("bearer "):
            token = token[7:]

        if not token:
            return None  # No token provided

        if token != expected_secret:
            return None  # Invalid token

        # Return basic authorization context for shared secret mode
        return AuthorizationContext(
            authenticated=True,
            principal="shared_secret_user",
            auth_method="shared_secret",
        )

    async def authorize(
        self,
        node: NodeLike,
        envelope: FameEnvelope,
        context: Optional[FameDeliveryContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Authorize access based on the current authentication context.

        Supports both new and legacy interfaces:
        - New: authorize(node, envelope, context: FameDeliveryContext)
        - Legacy: authorize(node, envelope, auth_context: AuthorizationContext)

        For shared secret auth, if authenticated then authorized.
        """
        # Backward compatibility: detect if context is actually an AuthorizationContext
        auth_context: Optional[AuthorizationContext] = None

        if context is None:
            auth_context = None
        elif hasattr(context, "authenticated"):
            # Legacy interface: context is an AuthorizationContext (duck typing check)
            auth_context = context  # type: ignore
        else:
            # New interface: context is a FameDeliveryContext
            if context and context.security and context.security.authorization:
                auth_context = context.security.authorization

        if not auth_context or not auth_context.authenticated:
            return None

        # For shared secret auth, if authenticated then authorized
        return auth_context

    async def validate_node_attach_request(
        self,
        node: NodeLike,
        frame: NodeAttachFrame,
        auth_context: Optional[AuthorizationContext] = None,
    ) -> Optional[AuthorizationContext]:
        """
        Validate node attach request using the already-authenticated context.
        No longer uses attach_token from frame - relies on wire-level auth.
        """
        if not auth_context or not auth_context.authenticated:
            return None  # Must be authenticated via wire-level auth first

        # The wire-level authentication already validated the credentials
        # Just create the node-specific authorization context
        return FameNodeAuthorizationContext(
            sub=frame.system_id,
            aud=node.id,
            instance_id=frame.instance_id,
            assigned_path=frame.assigned_path,
            accepted_capabilities=(list(frame.capabilities) if frame.capabilities else None),
            accepted_logicals=(list(frame.accepted_logicals) if frame.accepted_logicals else None),
        )



"""
Bearer token header authentication strategy implementation.
"""

from __future__ import annotations

import asyncio
from typing import Any, Optional

from naylence.fame.core import FameConnector
from naylence.fame.factory import create_resource
from naylence.fame.security.auth.auth_injection_strategy import AuthInjectionStrategy
from naylence.fame.security.auth.bearer_token_header_auth_injection_strategy_factory import (
    BearerTokenHeaderAuthInjectionStrategyConfig,
)
from naylence.fame.security.auth.token_provider import TokenProvider
from naylence.fame.security.auth.token_provider_factory import TokenProviderFactory
from naylence.fame.util.logging import getLogger

logger = getLogger(__name__)


class BearerTokenHeaderAuthInjectionStrategy(AuthInjectionStrategy):
    """Strategy for Bearer token in Authorization header."""

    def __init__(self, config: BearerTokenHeaderAuthInjectionStrategyConfig):
        self.config = config
        self._refresh_task: Optional[asyncio.Task] = None

    async def apply(self, connector: Any) -> None:
        """Apply authentication configuration to the connector."""
        # Create token provider
        token_provider = await create_resource(TokenProviderFactory, self.config.token_provider)

        # Set initial token
        await self._update_auth_header(connector, token_provider)

        # Start background refresh if needed
        self._start_refresh_task(connector, token_provider)

    async def _update_auth_header(self, connector: Any, token_provider: TokenProvider) -> None:
        """Update the connector's auth header with current token."""
        token = await token_provider.get_token()
        auth_header = f"Bearer {token.value}"

        # Use the connector's set_auth_header method if available
        if hasattr(connector, "set_auth_header"):
            getattr(connector, "set_auth_header")(auth_header)
        elif isinstance(connector, dict):
            connector["Authorization"] = auth_header
        else:
            logger.warning(f"Connector {type(connector)} doesn't support set_auth_header")

    def _start_refresh_task(self, connector: FameConnector, token_provider: TokenProvider) -> None:
        """Start background task to refresh token when needed."""

        async def refresh_loop():
            while True:
                try:
                    token = await token_provider.get_token()

                    # Calculate sleep time (refresh 30 seconds before expiry)
                    if token.expires_at:
                        import datetime

                        now = datetime.datetime.now(datetime.timezone.utc)
                        time_until_expiry = (token.expires_at - now).total_seconds()
                        sleep_time = max(time_until_expiry - 30, 60)  # At least 60 seconds
                    else:
                        sleep_time = 3600  # 1 hour default for tokens without expiry

                    await asyncio.sleep(sleep_time)

                    # Refresh the token
                    await self._update_auth_header(connector, token_provider)
                    logger.debug("auth_token_refreshed", connector_type=type(connector).__name__)

                except asyncio.CancelledError:
                    break
                except Exception as e:
                    logger.error("auth_token_refresh_failed", error=str(e), exc_info=True)
                    # Wait a bit before retrying
                    await asyncio.sleep(60)

        self._refresh_task = asyncio.create_task(refresh_loop())

    async def cleanup(self) -> None:
        """Clean up any background tasks or resources."""
        if self._refresh_task and not self._refresh_task.done():
            self._refresh_task.cancel()
            try:
                await self._refresh_task
            except asyncio.CancelledError:
                pass



from __future__ import annotations

from typing import Any, Callable, Dict, NamedTuple, Optional


# Lazy import cryptography and jwt only when needed
def _get_crypto_modules():
    """Get cryptography and jwt modules, importing them lazily."""
    try:
        from cryptography.hazmat.primitives import serialization
        from cryptography.hazmat.primitives.asymmetric import ec, ed448, ed25519, rsa
        from cryptography.hazmat.primitives.serialization import load_pem_public_key
        from jwt.algorithms import ECAlgorithm, OKPAlgorithm, RSAAlgorithm

        return {
            "serialization": serialization,
            "ec": ec,
            "ed448": ed448,
            "ed25519": ed25519,
            "rsa": rsa,
            "load_pem_public_key": load_pem_public_key,
            "ECAlgorithm": ECAlgorithm,
            "OKPAlgorithm": OKPAlgorithm,
            "RSAAlgorithm": RSAAlgorithm,
        }
    except ImportError:
        raise ImportError(
            "This functionality requires cryptography and pyjwt packages. "
            "Install with: pip install 'naylence-fame-runtime[crypto]'"
        )


def generate_key_pair_and_jwks(
    *,
    key_gen_fn: Callable[[], Any],
    algorithm: str,
    kid: str,
) -> dict[str, Any]:
    """
    :param key_gen_fn: zero-arg that returns a private_key object
    :param algorithm: e.g. "RS256", "ES256", "EdDSA"
    :param kid: key identifier
    :returns: dict with private_pem, public_pem, and jwks={"keys":[...]}
    """
    # Get crypto modules lazily
    crypto = _get_crypto_modules()
    serialization = crypto["serialization"]
    RSAAlgorithm = crypto["RSAAlgorithm"]
    ECAlgorithm = crypto["ECAlgorithm"]
    OKPAlgorithm = crypto["OKPAlgorithm"]

    # 1) generate private key
    priv = key_gen_fn()

    # 2) serialize private PEM
    private_pem = priv.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    ).decode()

    # 3) get public key object and PEM
    pub = priv.public_key()
    public_pem = pub.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    ).decode()

    # 4) pick correct algorithm class to call .to_jwk(...)
    alg = algorithm.upper()
    if alg.startswith("RS"):
        jwk_dict = RSAAlgorithm.to_jwk(pub, as_dict=True)
    elif alg.startswith("ES"):
        jwk_dict = ECAlgorithm.to_jwk(pub, as_dict=True)
    elif alg.startswith("ED"):
        jwk_dict = OKPAlgorithm.to_jwk(pub, as_dict=True)
    else:
        raise ValueError(f"Unsupported algorithm {algorithm!r}")

    # 5) enrich and return
    jwk_dict.update(
        {
            "kid": kid,
            "alg": algorithm,
            "use": "sig",
        }
    )
    return {
        "private_pem": private_pem,
        "public_pem": public_pem,
        "jwks": {"keys": [jwk_dict]},
    }


class DevKeyPair(NamedTuple):
    private_pem: str
    public_pem: str
    jwks: Dict[str, Any]


def detect_alg(pem: str) -> str:
    """
    Inspect a public-key PEM and return the matching JWT alg:
      - RSA keys    -> RS256
      - EC P-256    -> ES256
      - EC P-384    -> ES384
      - EC P-521    -> ES512
      - Ed25519     -> EdDSA
      - Ed448       -> EdDSA
    """
    # Get crypto modules lazily
    crypto = _get_crypto_modules()
    load_pem_public_key = crypto["load_pem_public_key"]
    rsa = crypto["rsa"]
    ec = crypto["ec"]
    ed25519 = crypto["ed25519"]
    ed448 = crypto["ed448"]

    key = load_pem_public_key(pem.encode())
    # RSA
    if isinstance(key, rsa.RSAPublicKey):
        return "RS256"
    # ECDSA
    if isinstance(key, ec.EllipticCurvePublicKey):
        curve = key.curve
        if isinstance(curve, ec.SECP256R1):
            return "ES256"
        if isinstance(curve, ec.SECP384R1):
            return "ES384"
        if isinstance(curve, ec.SECP521R1):
            return "ES512"
        raise ValueError(f"Unsupported EC curve: {curve.name}")
    # EdDSA
    if isinstance(key, ed25519.Ed25519PublicKey) or isinstance(key, ed448.Ed448PublicKey):
        return "EdDSA"
    raise ValueError("Could not detect algorithm from PEM")


def jwk_from_pem(
    pem: str,
    *,
    algorithm: Optional[str] = None,
    kid: str,
) -> Dict[str, Any]:
    """
    Convert a public-key PEM into a JWK dict, inferring or using the provided alg and tagging with kid.
    """
    # Get crypto modules lazily
    crypto = _get_crypto_modules()
    load_pem_public_key = crypto["load_pem_public_key"]
    RSAAlgorithm = crypto["RSAAlgorithm"]
    ECAlgorithm = crypto["ECAlgorithm"]
    OKPAlgorithm = crypto["OKPAlgorithm"]

    key_obj = load_pem_public_key(pem.encode())

    # determine algorithm
    alg = algorithm or detect_alg(pem)
    alg_upper = alg.upper()

    # build the JWK via PyJWT algorithm classes
    if alg_upper.startswith("RS"):
        jwk = RSAAlgorithm.to_jwk(key_obj, as_dict=True)  # type: ignore
    elif alg_upper.startswith("ES"):
        jwk = ECAlgorithm.to_jwk(key_obj, as_dict=True)  # type: ignore
    elif alg_upper.startswith("ED"):
        jwk = OKPAlgorithm.to_jwk(key_obj, as_dict=True)  # type: ignore
    else:
        raise ValueError(f"Unsupported JWK alg: {alg}")

    # enrich
    jwk.update(
        {
            "kid": kid,
            "alg": alg,
            "use": "sig",
        }
    )
    return jwk



"""
Authentication injection strategies for connectors.

These strategies know how to apply authentication configurations to connectors,
including setting up token providers and handling token refresh.
"""

from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from typing import Any, Optional

from naylence.fame.security.auth.auth_injection_strategy_factory import AuthInjectionStrategyConfig
from naylence.fame.util.logging import getLogger

logger = getLogger(__name__)


class AuthInjectionStrategy(ABC):
    """
    Base class for authentication injection strategies.

    Each strategy knows how to apply a specific type of Auth
    to a connector, including initial setup and ongoing token refresh.
    """

    def __init__(self, auth_config: AuthInjectionStrategyConfig):
        self.auth_config = auth_config
        self._refresh_task: Optional[asyncio.Task] = None

    @abstractmethod
    async def apply(self, connector: Any) -> None:
        """
        Apply authentication configuration to the connector.

        This should set up initial authentication and start any
        background refresh tasks if needed.
        """
        pass

    async def cleanup(self) -> None:
        """Clean up any background tasks or resources."""
        if self._refresh_task and not self._refresh_task.done():
            self._refresh_task.cancel()
            try:
                await self._refresh_task
            except asyncio.CancelledError:
                pass



from __future__ import annotations

from typing import Any, Optional

from naylence.fame.security.auth.auth_injection_strategy_factory import (
    AuthInjectionStrategyConfig,
    AuthInjectionStrategyFactory,
)

from .trace_emitter import TraceEmitter
from .trace_emitter_factory import TraceEmitterConfig, TraceEmitterFactory


class OpenTelemetryTraceEmitterConfig(TraceEmitterConfig):
    """Configuration for OpenTelemetryTraceEmitter."""

    type: str = "OpenTelemetryTraceEmitter"

    service_name: str = "naylence-service"
    endpoint: Optional[str] = None
    environment: Optional[str] = None
    sampler: Optional[str] = None
    headers: Optional[dict[str, str]] = None

    auth: Optional[AuthInjectionStrategyConfig] = None


class OpenTelemetryTraceEmitterFactory(TraceEmitterFactory):
    """Factory for creating OpenTelemetryTraceEmitter instances."""

    type: str = "OpenTelemetryTraceEmitter"

    async def create(
        self,
        config: Optional[OpenTelemetryTraceEmitterConfig | dict[str, Any]] = None,
        service_name: Optional[str] = None,
        endpoint: Optional[str] = None,
        environment: Optional[str] = None,
        sampler: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
        tracer: Optional[Any] = None,
        **kwargs: Any,
    ) -> TraceEmitter:
        """Create an OpenTelemetryTraceEmitter instance."""
        from .open_telemetry_trace_emitter import OpenTelemetryTraceEmitter

        if config is None:
            config = OpenTelemetryTraceEmitterConfig()
        elif isinstance(config, dict):
            config = OpenTelemetryTraceEmitterConfig(**config)

        service_name = service_name or config.service_name
        tracer = kwargs.get("tracer")

        merged_headers = {**(config.headers or {}), **(headers or {})}

        if config.auth:
            auth_strategy = await AuthInjectionStrategyFactory.create_auth_strategy(config.auth)
            # Set a default "Authorization" header on the dictionary
            await auth_strategy.apply(merged_headers)

        from .otel_setup import setup_otel

        setup_otel(
            service_name=service_name,
            endpoint=endpoint or config.endpoint,
            environment=environment or config.environment,
            sampler=sampler or config.sampler or "parentbased_always_on",
            headers=merged_headers if merged_headers else None,
        )

        return OpenTelemetryTraceEmitter(service_name=service_name, tracer=tracer)



from __future__ import annotations

from typing import Any, Dict, Optional

from pydantic import Field

from naylence.fame.factory import ResourceConfig, ResourceFactory
from naylence.fame.sentinel.store.route_store import RouteEntry, RouteStore
from naylence.fame.storage.in_memory_key_value_store import InMemoryKVStore


class RouteStoreConfig(ResourceConfig):
    type: str = "RouteStore"
    params: Optional[Dict[str, Any]] = Field(
        default=None,
        description="Implementation-specific parameters (e.g. Redis URL)",
    )


class RouteStoreFactory(ResourceFactory[RouteStore, RouteStoreConfig]): ...


class InMemoryRouteStoreFactory(RouteStoreFactory):
    async def create(
        self, config: Optional[RouteStoreConfig | dict[str, Any]] = None, **kwargs: Any
    ) -> RouteStore:
        # 'none' store: no-op implementation
        if not config:
            raise ValueError("Config not set")

        return InMemoryKVStore(RouteEntry)
